"""
References:
1. https://docs.aws.amazon.com/lambda/latest/dg/services-alb.html
2. https://docs.aws.amazon.com/elasticloadbalancing/latest/application/lambda-functions.html  # noqa: E501
"""

from __future__ import annotations

import pytest

from mangum import Mangum
from mangum.handlers.alb import ALB


def get_mock_aws_alb_event(
    method,
    path,
    query_parameters: dict[str, list[str]] | None,
    headers: dict[str, list[str]] | None,
    body,
    body_base64_encoded,
    multi_value_headers: bool,
):
    """Return a mock AWS ELB event.

    The `query_parameters` parameter must be given in the
    `multiValueQueryStringParameters` format - and if `multi_value_headers`
    is disabled, then they are simply transformed in to the
    `queryStringParameters` format.
    Similarly for `headers`.
    If `headers` is None, then some defaults will be used.
    if `query_parameters` is None, then no query parameters will be used.
    """
    resp = {
        "requestContext": {
            "elb": {
                "targetGroupArn": (
                    "arn:aws:elasticloadbalancing:us-east-2:123456789012:"
                    "targetgroup/lambda-279XGJDqGZ5rsrHC2Fjr/49e9d65c45c6791a"
                )
            }
        },
        "httpMethod": method,
        "path": path,
        "body": body,
        "isBase64Encoded": body_base64_encoded,
    }

    if headers is None:
        headers = {
            "accept": ["text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8"],
            "accept-encoding": ["gzip"],
            "accept-language": ["en-US,en;q=0.9"],
            "connection": ["keep-alive"],
            "host": ["lambda-alb-123578498.us-east-2.elb.amazonaws.com"],
            "upgrade-insecure-requests": ["1"],
            "user-agent": [
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 "
                "(KHTML, like Gecko) Chrome/71.0.3578.98 Safari/537.36"
            ],
            "x-amzn-trace-id": ["Root=1-5c536348-3d683b8b04734faae651f476"],
            "x-forwarded-for": ["72.12.164.125"],
            "x-forwarded-port": ["80"],
            "x-forwarded-proto": ["http"],
            "x-imforwards": ["20"],
        }

    query_parameters = {} if query_parameters is None else query_parameters

    # Only set one of `queryStringParameters`/`multiValueQueryStringParameters`
    # and one of `headers`/multiValueHeaders (per AWS docs for ALB/lambda)
    if multi_value_headers:
        resp["multiValueQueryStringParameters"] = query_parameters
        resp["multiValueHeaders"] = headers
    else:
        # Take the last query parameter/cookie (per AWS docs for ALB/lambda)
        resp["queryStringParameters"] = {k: (v[-1] if len(v) > 0 else []) for k, v in query_parameters.items()}
        resp["headers"] = {k: (v[-1] if len(v) > 0 else []) for k, v in headers.items()}

    return resp


@pytest.mark.parametrize(
    "method,path,query_parameters,headers,req_body,body_base64_encoded,query_string,scope_body,multi_value_headers",
    [
        ("GET", "/hello/world", None, None, None, False, b"", None, False),
        (
            "GET",
            "/lambda",
            {
                "q1": ["1234ABCD"],
                "q2": ["b+c"],  # not encoded
                "q3": ["b%20c"],  # encoded
                "q4": ["/some/path/"],  # not encoded
                "q5": ["%2Fsome%2Fpath%2F"],  # encoded
            },
            None,
            "",
            False,
            b"q1=1234ABCD&q2=b+c&q3=b+c&q4=%2Fsome%2Fpath%2F&q5=%2Fsome%2Fpath%2F",
            "",
            False,
        ),
        (
            "POST",
            "/",
            {"name": ["me"]},
            None,
            "field1=value1&field2=value2",
            False,
            b"name=me",
            b"field1=value1&field2=value2",
            False,
        ),
        # Duplicate query params with multi-value headers disabled:
        (
            "POST",
            "/",
            {"name": ["me", "you"]},
            None,
            None,
            False,
            b"name=you",
            None,
            False,
        ),
        #  Duplicate query params with multi-value headers enable:
        (
            "GET",
            "/my/resource",
            {"name": ["me", "you"]},
            None,
            None,
            False,
            b"name=me&name=you",
            None,
            True,
        ),
        (
            "GET",
            "",
            {"name": ["me", "you"], "pet": ["dog"]},
            None,
            None,
            False,
            b"name=me&name=you&pet=dog",
            None,
            True,
        ),
        # A 1x1 red px gif
        (
            "POST",
            "/img",
            None,
            None,
            b"R0lGODdhAQABAIABAP8AAAAAACwAAAAAAQABAAACAkQBADs=",
            True,
            b"",
            b"GIF87a\x01\x00\x01\x00\x80\x01\x00\xff\x00\x00\x00\x00\x00,"
            b"\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02D\x01\x00;",
            False,
        ),
        (
            "POST",
            "/form-submit",
            None,
            None,
            b"say=Hi&to=Mom",
            False,
            b"",
            b"say=Hi&to=Mom",
            False,
        ),
    ],
)
def test_aws_alb_scope_real(
    method,
    path,
    query_parameters,
    headers,
    req_body,
    body_base64_encoded,
    query_string,
    scope_body,
    multi_value_headers,
):
    event = get_mock_aws_alb_event(
        method,
        path,
        query_parameters,
        headers,
        req_body,
        body_base64_encoded,
        multi_value_headers,
    )
    example_context = {}
    handler = ALB(event, example_context, {"api_gateway_base_path": "/"})

    scope_path = path
    if scope_path == "":
        scope_path = "/"

    assert isinstance(handler.body, bytes)
    assert handler.scope == {
        "asgi": {"version": "3.0", "spec_version": "2.0"},
        "aws.context": {},
        "aws.event": event,
        "client": ("72.12.164.125", 0),
        "headers": [
            [
                b"accept",
                b"text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8",
            ],
            [b"accept-encoding", b"gzip"],
            [b"accept-language", b"en-US,en;q=0.9"],
            [b"connection", b"keep-alive"],
            [b"host", b"lambda-alb-123578498.us-east-2.elb.amazonaws.com"],
            [b"upgrade-insecure-requests", b"1"],
            [
                b"user-agent",
                b"Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36"
                b" (KHTML, like Gecko) Chrome/71.0.3578.98 Safari/537.36",
            ],
            [b"x-amzn-trace-id", b"Root=1-5c536348-3d683b8b04734faae651f476"],
            [b"x-forwarded-for", b"72.12.164.125"],
            [b"x-forwarded-port", b"80"],
            [b"x-forwarded-proto", b"http"],
            [b"x-imforwards", b"20"],
        ],
        "http_version": "1.1",
        "method": method,
        "path": scope_path,
        "query_string": query_string,
        "raw_path": None,
        "root_path": "",
        "scheme": "http",
        "server": ("lambda-alb-123578498.us-east-2.elb.amazonaws.com", 80),
        "type": "http",
    }

    if handler.body:
        assert handler.body == scope_body
    else:
        assert handler.body == b""


@pytest.mark.parametrize("multi_value_headers_enabled", (True, False))
def test_aws_alb_set_cookies(multi_value_headers_enabled) -> None:
    async def app(scope, receive, send):
        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": [
                    [b"content-type", b"text/plain; charset=utf-8"],
                    [b"set-cookie", b"cookie1=cookie1; Secure"],
                    [b"set-cookie", b"cookie2=cookie2; Secure"],
                ],
            }
        )
        await send({"type": "http.response.body", "body": b"Hello, world!"})

    handler = Mangum(app, lifespan="off")
    event = get_mock_aws_alb_event("GET", "/test", {}, None, None, False, multi_value_headers_enabled)
    response = handler(event, {})

    expected_response = {
        "statusCode": 200,
        "isBase64Encoded": False,
        "body": "Hello, world!",
    }
    if multi_value_headers_enabled:
        expected_response["multiValueHeaders"] = {
            "set-cookie": ["cookie1=cookie1; Secure", "cookie2=cookie2; Secure"],
            "content-type": ["text/plain; charset=utf-8"],
        }
    else:
        expected_response["headers"] = {
            "content-type": "text/plain; charset=utf-8",
            # Should see case mutated keys to avoid duplicate keys:
            "set-cookie": "cookie1=cookie1; Secure",
            "Set-cookie": "cookie2=cookie2; Secure",
        }
    assert response == expected_response


@pytest.mark.parametrize(
    "method,content_type,raw_res_body,res_body,res_base64_encoded",
    [
        ("GET", b"text/plain; charset=utf-8", b"Hello world", "Hello world", False),
        # A 1x1 red px gif
        (
            "POST",
            b"image/gif",
            b"GIF87a\x01\x00\x01\x00\x80\x01\x00\xff\x00\x00\x00\x00\x00,"
            b"\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02D\x01\x00;",
            "R0lGODdhAQABAIABAP8AAAAAACwAAAAAAQABAAACAkQBADs=",
            True,
        ),
    ],
)
def test_aws_alb_response(method, content_type, raw_res_body, res_body, res_base64_encoded):
    async def app(scope, receive, send):
        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": [[b"content-type", content_type]],
            }
        )
        await send({"type": "http.response.body", "body": raw_res_body})

    event = get_mock_aws_alb_event(method, "/test", {}, None, None, False, False)

    handler = Mangum(app, lifespan="off")

    response = handler(event, {})
    assert response == {
        "statusCode": 200,
        "isBase64Encoded": res_base64_encoded,
        "headers": {"content-type": content_type.decode()},
        "body": res_body,
    }


def test_aws_alb_response_extra_mime_types():
    content_type = b"application/x-yaml"
    utf_res_body = "name: 'John Doe'"
    raw_res_body = utf_res_body.encode()
    b64_res_body = "bmFtZTogJ0pvaG4gRG9lJw=="

    async def app(scope, receive, send):
        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": [[b"content-type", content_type]],
            }
        )
        await send({"type": "http.response.body", "body": raw_res_body})

    event = get_mock_aws_alb_event("GET", "/test", {}, None, None, False, False)

    # Test default behavior
    handler = Mangum(app, lifespan="off")
    response = handler(event, {})
    assert content_type.decode() not in handler.config["text_mime_types"]
    assert response == {
        "statusCode": 200,
        "isBase64Encoded": True,
        "headers": {"content-type": content_type.decode()},
        "body": b64_res_body,
    }

    # Test with modified text mime types
    handler = Mangum(app, lifespan="off")
    handler.config["text_mime_types"].append(content_type.decode())
    response = handler(event, {})
    assert response == {
        "statusCode": 200,
        "isBase64Encoded": False,
        "headers": {"content-type": content_type.decode()},
        "body": utf_res_body,
    }


@pytest.mark.parametrize("multi_value_headers_enabled", (True, False))
def test_aws_alb_exclude_headers(multi_value_headers_enabled) -> None:
    async def app(scope, receive, send):
        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": [
                    [b"content-type", b"text/plain; charset=utf-8"],
                    [b"x-custom-header", b"test"],
                ],
            }
        )
        await send({"type": "http.response.body", "body": b"Hello, world!"})

    handler = Mangum(app, lifespan="off", exclude_headers=["x-custom-header"])
    event = get_mock_aws_alb_event("GET", "/test", {}, None, None, False, multi_value_headers_enabled)
    response = handler(event, {})

    expected_response = {
        "statusCode": 200,
        "isBase64Encoded": False,
        "body": "Hello, world!",
    }
    if multi_value_headers_enabled:
        expected_response["multiValueHeaders"] = {
            "content-type": ["text/plain; charset=utf-8"],
        }
    else:
        expected_response["headers"] = {
            "content-type": "text/plain; charset=utf-8",
        }
    assert response == expected_response



from __future__ import annotations

import os
from typing import Any

import pytest


@pytest.fixture
def mock_aws_api_gateway_event(request: pytest.FixtureRequest):
    method = request.param[0]
    body = request.param[1]
    multi_value_query_parameters = request.param[2]
    event: dict[str, Any] = {
        "path": "/test/hello",
        "body": body,
        "headers": {
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
            "Accept-Encoding": "gzip, deflate, lzma, sdch, br",
            "Accept-Language": "en-US,en;q=0.8",
            "CloudFront-Forwarded-Proto": "https",
            "CloudFront-Is-Desktop-Viewer": "true",
            "CloudFront-Is-Mobile-Viewer": "false",
            "CloudFront-Is-SmartTV-Viewer": "false",
            "CloudFront-Is-Tablet-Viewer": "false",
            "CloudFront-Viewer-Country": "US",
            "Cookie": "cookie1; cookie2",
            "Host": "test.execute-api.us-west-2.amazonaws.com",
            "Upgrade-Insecure-Requests": "1",
            "X-Forwarded-For": "192.168.100.1, 192.168.1.1",
            "X-Forwarded-Port": "443",
            "X-Forwarded-Proto": "https",
        },
        "pathParameters": {"proxy": "hello"},
        "requestContext": {
            "accountId": "123456789012",
            "resourceId": "us4z18",
            "stage": "Prod",
            "requestId": "41b45ea3-70b5-11e6-b7bd-69b5aaebc7d9",
            "identity": {
                "cognitoIdentityPoolId": "",
                "accountId": "",
                "cognitoIdentityId": "",
                "caller": "",
                "apiKey": "",
                "sourceIp": "192.168.100.1",
                "cognitoAuthenticationType": "",
                "cognitoAuthenticationProvider": "",
                "userArn": "",
                "userAgent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/52.0.2743.82 Safari/537.36 OPR/39.0.2256.48",  # noqa: E501
                "user": "",
            },
            "resourcePath": "/{proxy+}",
            "httpMethod": method,
            "apiId": "123",
        },
        "resource": "/{proxy+}",
        "httpMethod": method,
        "queryStringParameters": (
            {k: v[0] for k, v in multi_value_query_parameters.items()} if multi_value_query_parameters else None
        ),
        "multiValueQueryStringParameters": multi_value_query_parameters or None,
        "stageVariables": {"stageVarName": "stageVarValue"},
    }
    return event


@pytest.fixture
def mock_http_api_event_v2(request: pytest.FixtureRequest):
    method = request.param[0]
    body = request.param[1]
    multi_value_query_parameters = request.param[2]
    query_string = request.param[3]
    event = {
        "version": "2.0",
        "routeKey": "$default",
        "rawPath": "/my/path",
        "rawQueryString": query_string,
        "cookies": ["cookie1", "cookie2"],
        "headers": {
            "accept-encoding": "gzip,deflate",
            "x-forwarded-port": "443",
            "x-forwarded-proto": "https",
            "host": "test.execute-api.us-west-2.amazonaws.com",
        },
        "queryStringParameters": (
            {k: v[0] for k, v in multi_value_query_parameters.items()} if multi_value_query_parameters else None
        ),
        "requestContext": {
            "accountId": "123456789012",
            "apiId": "api-id",
            "authorizer": {
                "jwt": {
                    "claims": {"claim1": "value1", "claim2": "value2"},
                    "scopes": ["scope1", "scope2"],
                }
            },
            "domainName": "id.execute-api.us-east-1.amazonaws.com",
            "domainPrefix": "id",
            "http": {
                "method": method,
                "path": "/my/path",
                "protocol": "HTTP/1.1",
                "sourceIp": "192.168.100.1",
                "userAgent": "agent",
            },
            "requestId": "id",
            "routeKey": "$default",
            "stage": "$default",
            "time": "12/Mar/2020:19:03:58 +0000",
            "timeEpoch": 1583348638390,
        },
        "body": body,
        "pathParameters": {"parameter1": "value1"},
        "isBase64Encoded": False,
        "stageVariables": {"stageVariable1": "value1", "stageVariable2": "value2"},
    }

    return event


@pytest.fixture
def mock_http_api_event_v1(request: pytest.FixtureRequest):
    method = request.param[0]
    body = request.param[1]
    multi_value_query_parameters = request.param[2]
    query_string = request.param[3]
    event = {
        "version": "1.0",
        "routeKey": "$default",
        "rawPath": "/my/path",
        "path": "/my/path",
        "httpMethod": method,
        "rawQueryString": query_string,
        "cookies": ["cookie1", "cookie2"],
        "headers": {
            "accept-encoding": "gzip,deflate",
            "x-forwarded-port": "443",
            "x-forwarded-proto": "https",
            "host": "test.execute-api.us-west-2.amazonaws.com",
        },
        "queryStringParameters": (
            {k: v[-1] for k, v in multi_value_query_parameters.items()} if multi_value_query_parameters else None
        ),
        "multiValueQueryStringParameters": (
            {k: v for k, v in multi_value_query_parameters.items()} if multi_value_query_parameters else None
        ),
        "requestContext": {
            "accountId": "123456789012",
            "apiId": "api-id",
            "authorizer": {
                "jwt": {
                    "claims": {"claim1": "value1", "claim2": "value2"},
                    "scopes": ["scope1", "scope2"],
                }
            },
            "domainName": "id.execute-api.us-east-1.amazonaws.com",
            "domainPrefix": "id",
            "http": {
                "protocol": "HTTP/1.1",
                "sourceIp": "192.168.100.1",
                "userAgent": "agent",
            },
            "requestId": "id",
            "routeKey": "$default",
            "stage": "$default",
            "time": "12/Mar/2020:19:03:58 +0000",
            "timeEpoch": 1583348638390,
        },
        "body": body,
        "pathParameters": {"parameter1": "value1"},
        "isBase64Encoded": False,
        "stageVariables": {"stageVariable1": "value1", "stageVariable2": "value2"},
    }

    return event


@pytest.fixture
def mock_lambda_at_edge_event(request: pytest.FixtureRequest):
    method = request.param[0]
    path = request.param[1]
    query_string = request.param[2]
    body = request.param[3]

    headers_raw = {
        "accept-encoding": "gzip,deflate",
        "x-forwarded-port": "443",
        "x-forwarded-for": "192.168.100.1",
        "x-forwarded-proto": "https",
        "host": "test.execute-api.us-west-2.amazonaws.com",
    }
    headers = {}
    for key, value in headers_raw.items():
        headers[key.lower()] = [{"key": key, "value": value}]

    event = {
        "Records": [
            {
                "cf": {
                    "config": {
                        "distributionDomainName": "mock-distribution.local.localhost",
                        "distributionId": "ABC123DEF456G",
                        "eventType": "origin-request",
                        "requestId": "lBEBo2N0JKYUP2JXwn_4am2xAXB2GzcL2FlwXI8G59PA8wghF2ImFQ==",
                    },
                    "request": {
                        "clientIp": "192.168.100.1",
                        "headers": headers,
                        "method": method,
                        "origin": {
                            "custom": {
                                "customHeaders": {
                                    "x-lae-env-custom-var": [
                                        {
                                            "key": "x-lae-env-custom-var",
                                            "value": "environment variable",
                                        }
                                    ],
                                },
                                "domainName": "www.example.com",
                                "keepaliveTimeout": 5,
                                "path": "",
                                "port": 80,
                                "protocol": "http",
                                "readTimeout": 30,
                                "sslProtocols": ["TLSv1", "TLSv1.1", "TLSv1.2"],
                            }
                        },
                        "querystring": query_string,
                        "uri": path,
                    },
                }
            }
        ]
    }

    if body is not None:
        event["Records"][0]["cf"]["request"]["body"] = {
            "inputTruncated": False,
            "action": "read-only",
            "encoding": "text",
            "data": body,
        }

    return dict(method=method, path=path, query_string=query_string, body=body, event=event)


@pytest.fixture(scope="session", autouse=True)
def aws_credentials():
    """Mocked AWS Credentials for moto."""
    os.environ["AWS_DEFAULT_REGION"] = "testing"
    os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"
    os.environ["AWS_SECURITY_TOKEN"] = "testing"
    os.environ["AWS_SESSION_TOKEN"] = "testing"



import urllib.parse

import pytest

from mangum import Mangum
from mangum.handlers.lambda_at_edge import LambdaAtEdge


def mock_lambda_at_edge_event(method, path, multi_value_query_parameters, body, body_base64_encoded):
    headers_raw = {
        "accept-encoding": "gzip,deflate",
        "x-forwarded-port": "443",
        "x-forwarded-for": "192.168.100.1",
        "x-forwarded-proto": "https",
        "host": "test.execute-api.us-west-2.amazonaws.com",
    }
    headers = {}
    for key, value in headers_raw.items():
        headers[key.lower()] = [{"key": key, "value": value}]

    event = {
        "Records": [
            {
                "cf": {
                    "config": {
                        "distributionDomainName": "mock-distribution.local.localhost",
                        "distributionId": "ABC123DEF456G",
                        "eventType": "origin-request",
                        "requestId": "lBEBo2N0JKYUP2JXwn_4am2xAXB2GzcL2FlwXI8G59PA8wghF2ImFQ==",
                    },
                    "request": {
                        "clientIp": "192.168.100.1",
                        "headers": headers,
                        "method": method,
                        "origin": {
                            "custom": {
                                "customHeaders": {
                                    "x-lae-env-custom-var": [
                                        {
                                            "key": "x-lae-env-custom-var",
                                            "value": "environment variable",
                                        }
                                    ],
                                },
                                "domainName": "www.example.com",
                                "keepaliveTimeout": 5,
                                "path": "",
                                "port": 80,
                                "protocol": "http",
                                "readTimeout": 30,
                                "sslProtocols": ["TLSv1", "TLSv1.1", "TLSv1.2"],
                            }
                        },
                        "querystring": urllib.parse.urlencode(
                            (multi_value_query_parameters if multi_value_query_parameters else {}),
                            doseq=True,
                        ),
                        "uri": path,
                    },
                }
            }
        ]
    }

    if body is not None:
        event["Records"][0]["cf"]["request"]["body"] = {
            "inputTruncated": False,
            "action": "read-only",
            "encoding": "base64" if body_base64_encoded else "text",
            "data": body,
        }
    return event


def test_aws_cf_lambda_at_edge_scope_basic():
    """
    Test the event from the AWS docs
    """
    example_event = {
        "Records": [
            {
                "cf": {
                    "config": {
                        "distributionDomainName": "d111111abcdef8.cloudfront.net",
                        "distributionId": "EDFDVBD6EXAMPLE",
                        "eventType": "origin-request",
                        "requestId": "4TyzHTaYWb1GX1qTfsHhEqV6HUDd_BzoBZnwfnvQc_1oF26ClkoUSEQ==",
                    },
                    "request": {
                        "clientIp": "203.0.113.178",
                        "headers": {
                            "x-forwarded-for": [{"key": "X-Forwarded-For", "value": "203.0.113.178"}],
                            "user-agent": [{"key": "User-Agent", "value": "Amazon CloudFront"}],
                            "via": [
                                {
                                    "key": "Via",
                                    "value": "2.0 2afae0d44e2540f472c0635ab62c232b.cloudfront.net (CloudFront)",
                                }
                            ],
                            "host": [{"key": "Host", "value": "example.org"}],
                            "cache-control": [
                                {
                                    "key": "Cache-Control",
                                    "value": "no-cache, cf-no-cache",
                                }
                            ],
                        },
                        "method": "GET",
                        "origin": {
                            "custom": {
                                "customHeaders": {},
                                "domainName": "example.org",
                                "keepaliveTimeout": 5,
                                "path": "",
                                "port": 443,
                                "protocol": "https",
                                "readTimeout": 30,
                                "sslProtocols": ["TLSv1", "TLSv1.1", "TLSv1.2"],
                            }
                        },
                        "querystring": "",
                        "uri": "/",
                    },
                }
            }
        ]
    }
    example_context = {}
    handler = LambdaAtEdge(example_event, example_context, {"api_gateway_base_path": "/"})

    assert isinstance(handler.body, bytes)
    assert handler.scope == {
        "asgi": {"version": "3.0", "spec_version": "2.0"},
        "aws.context": {},
        "aws.event": example_event,
        "client": ("203.0.113.178", 0),
        "headers": [
            [b"x-forwarded-for", b"203.0.113.178"],
            [b"user-agent", b"Amazon CloudFront"],
            [
                b"via",
                b"2.0 2afae0d44e2540f472c0635ab62c232b.cloudfront.net (CloudFront)",
            ],
            [b"host", b"example.org"],
            [b"cache-control", b"no-cache, cf-no-cache"],
        ],
        "http_version": "1.1",
        "method": "GET",
        "path": "/",
        "query_string": b"",
        "raw_path": None,
        "root_path": "",
        "scheme": "https",
        "server": ("example.org", 80),
        "type": "http",
    }


@pytest.mark.parametrize(
    "method,path,multi_value_query_parameters,req_body,body_base64_encoded,query_string,scope_body",
    [
        ("GET", "/hello/world", None, None, False, b"", None),
        (
            "POST",
            "/",
            {"name": ["me"]},
            "field1=value1&field2=value2",
            False,
            b"name=me",
            b"field1=value1&field2=value2",
        ),
        (
            "GET",
            "/my/resource",
            {"name": ["me", "you"]},
            None,
            False,
            b"name=me&name=you",
            None,
        ),
        (
            "GET",
            "",
            {"name": ["me", "you"], "pet": ["dog"]},
            None,
            False,
            b"name=me&name=you&pet=dog",
            None,
        ),
        # A 1x1 red px gif
        (
            "POST",
            "/img",
            None,
            b"R0lGODdhAQABAIABAP8AAAAAACwAAAAAAQABAAACAkQBADs=",
            True,
            b"",
            b"GIF87a\x01\x00\x01\x00\x80\x01\x00\xff\x00\x00\x00\x00\x00,"
            b"\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02D\x01\x00;",
        ),
        ("POST", "/form-submit", None, b"say=Hi&to=Mom", False, b"", b"say=Hi&to=Mom"),
    ],
)
def test_aws_api_gateway_scope_real(
    method,
    path,
    multi_value_query_parameters,
    req_body,
    body_base64_encoded,
    query_string,
    scope_body,
):
    event = mock_lambda_at_edge_event(method, path, multi_value_query_parameters, req_body, body_base64_encoded)
    example_context = {}
    handler = LambdaAtEdge(event, example_context, {"api_gateway_base_path": "/"})

    assert handler.scope == {
        "asgi": {"version": "3.0", "spec_version": "2.0"},
        "aws.context": {},
        "aws.event": event,
        "client": ("192.168.100.1", 0),
        "headers": [
            [b"accept-encoding", b"gzip,deflate"],
            [b"x-forwarded-port", b"443"],
            [b"x-forwarded-for", b"192.168.100.1"],
            [b"x-forwarded-proto", b"https"],
            [b"host", b"test.execute-api.us-west-2.amazonaws.com"],
        ],
        "http_version": "1.1",
        "method": method,
        "path": path,
        "query_string": query_string,
        "raw_path": None,
        "root_path": "",
        "scheme": "https",
        "server": ("test.execute-api.us-west-2.amazonaws.com", 443),
        "type": "http",
    }

    if handler.body:
        assert handler.body == scope_body
    else:
        assert handler.body == b""


@pytest.mark.parametrize(
    "method,content_type,raw_res_body,res_body,res_base64_encoded",
    [
        ("GET", b"text/plain; charset=utf-8", b"Hello world", "Hello world", False),
        # A 1x1 red px gif
        (
            "POST",
            b"image/gif",
            b"GIF87a\x01\x00\x01\x00\x80\x01\x00\xff\x00\x00\x00\x00\x00,"
            b"\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02D\x01\x00;",
            "R0lGODdhAQABAIABAP8AAAAAACwAAAAAAQABAAACAkQBADs=",
            True,
        ),
    ],
)
def test_aws_lambda_at_edge_response(method, content_type, raw_res_body, res_body, res_base64_encoded):
    async def app(scope, receive, send):
        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": [[b"content-type", content_type]],
            }
        )
        await send({"type": "http.response.body", "body": raw_res_body})

    event = mock_lambda_at_edge_event(method, "/test", {}, None, False)

    handler = Mangum(app, lifespan="off")

    response = handler(event, {})
    assert response == {
        "status": 200,
        "isBase64Encoded": res_base64_encoded,
        "headers": {"content-type": [{"key": "content-type", "value": content_type.decode()}]},
        "body": res_body,
    }


def test_aws_lambda_at_edge_response_extra_mime_types():
    content_type = b"application/x-yaml"
    utf_res_body = "name: 'John Doe'"
    raw_res_body = utf_res_body.encode()
    b64_res_body = "bmFtZTogJ0pvaG4gRG9lJw=="

    async def app(scope, receive, send):
        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": [[b"content-type", content_type]],
            }
        )
        await send({"type": "http.response.body", "body": raw_res_body})

    event = mock_lambda_at_edge_event("POST", "/test", {}, None, False)

    # Test default behavior
    handler = Mangum(app, lifespan="off")
    response = handler(event, {})
    assert content_type.decode() not in handler.config["text_mime_types"]
    assert response == {
        "status": 200,
        "isBase64Encoded": True,
        "headers": {"content-type": [{"key": "content-type", "value": content_type.decode()}]},
        "body": b64_res_body,
    }

    # Test with modified text mime types
    handler = Mangum(app, lifespan="off")
    handler.config["text_mime_types"].append(content_type.decode())
    response = handler(event, {})
    assert response == {
        "status": 200,
        "isBase64Encoded": False,
        "headers": {"content-type": [{"key": "content-type", "value": content_type.decode()}]},
        "body": utf_res_body,
    }


def test_aws_lambda_at_edge_exclude_():
    async def app(scope, receive, send):
        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": [
                    [b"content-type", b"text/plain; charset=utf-8"],
                    [b"x-custom-header", b"test"],
                ],
            }
        )
        await send({"type": "http.response.body", "body": b"Hello world"})

    event = mock_lambda_at_edge_event("GET", "/test", {}, None, False)

    handler = Mangum(app, lifespan="off", exclude_headers=["x-custom-header"])

    response = handler(event, {})
    assert response == {
        "status": 200,
        "isBase64Encoded": False,
        "headers": {"content-type": [{"key": "content-type", "value": b"text/plain; charset=utf-8".decode()}]},
        "body": "Hello world",
    }



import urllib.parse

import pytest

from mangum import Mangum
from mangum.handlers.api_gateway import APIGateway


def get_mock_aws_api_gateway_event(method, path, multi_value_query_parameters, body, body_base64_encoded):
    return {
        "path": path,
        "body": body,
        "isBase64Encoded": body_base64_encoded,
        "headers": {
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
            "Accept-Encoding": "gzip, deflate, lzma, sdch, br",
            "Accept-Language": "en-US,en;q=0.8",
            "CloudFront-Forwarded-Proto": "https",
            "CloudFront-Is-Desktop-Viewer": "true",
            "CloudFront-Is-Mobile-Viewer": "false",
            "CloudFront-Is-SmartTV-Viewer": "false",
            "CloudFront-Is-Tablet-Viewer": "false",
            "CloudFront-Viewer-Country": "US",
            "Cookie": "cookie1; cookie2",
            "Host": "test.execute-api.us-west-2.amazonaws.com",
            "Upgrade-Insecure-Requests": "1",
            "X-Forwarded-For": "192.168.100.1, 192.168.1.1",
            "X-Forwarded-Port": "443",
            "X-Forwarded-Proto": "https",
        },
        "pathParameters": {"proxy": "hello"},
        "requestContext": {
            "accountId": "123456789012",
            "resourceId": "us4z18",
            "stage": "Prod",
            "requestId": "41b45ea3-70b5-11e6-b7bd-69b5aaebc7d9",
            "identity": {
                "cognitoIdentityPoolId": "",
                "accountId": "",
                "cognitoIdentityId": "",
                "caller": "",
                "apiKey": "",
                "sourceIp": "192.168.100.1",
                "cognitoAuthenticationType": "",
                "cognitoAuthenticationProvider": "",
                "userArn": "",
                "userAgent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_6) "
                "AppleWebKit/537.36 (KHTML, like Gecko) "
                "Chrome/52.0.2743.82 Safari/537.36 OPR/39.0.2256.48",
                "user": "",
            },
            "resourcePath": "/{proxy+}",
            "httpMethod": method,
            "apiId": "123",
        },
        "resource": "/{proxy+}",
        "httpMethod": method,
        "multiValueQueryStringParameters": (
            {k: v for k, v in multi_value_query_parameters.items()} if multi_value_query_parameters else None
        ),
        "stageVariables": {"stageVarName": "stageVarValue"},
    }


def test_aws_api_gateway_scope_basic():
    """
    Test the event from the AWS docs
    """
    example_event = {
        "resource": "/",
        "path": "/",
        "httpMethod": "GET",
        "requestContext": {"resourcePath": "/", "httpMethod": "GET", "path": "/Prod/"},
        "headers": {
            "accept": "text/html,application/xhtml+xml,application/xml;"
            "q=0.9,image/webp,image/apng,*/*;q=0.8,"
            "application/signed-exchange;v=b3;q=0.9",
            "accept-encoding": "gzip, deflate, br",
            "Host": "70ixmpl4fl.execute-api.us-east-2.amazonaws.com",
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
            "AppleWebKit/537.36 (KHTML, like Gecko) "
            "Chrome/80.0.3987.132 Safari/537.36",
            "X-Amzn-Trace-Id": "Root=1-5e66d96f-7491f09xmpl79d18acf3d050",
        },
        "multiValueHeaders": {
            "accept": [
                "text/html,application/xhtml+xml,application/xml;"
                "q=0.9,image/webp,image/apng,*/*;q=0.8,"
                "application/signed-exchange;v=b3;q=0.9"
            ],
            "accept-encoding": ["gzip, deflate, br"],
        },
        "queryStringParameters": {"foo": "bar"},
        "multiValueQueryStringParameters": None,
        "pathParameters": None,
        "stageVariables": None,
        "body": None,
        "isBase64Encoded": False,
    }
    example_context = {}
    handler = APIGateway(example_event, example_context, {"api_gateway_base_path": "/"})

    assert isinstance(handler.body, bytes)
    assert handler.scope == {
        "asgi": {"version": "3.0", "spec_version": "2.0"},
        "aws.context": {},
        "aws.event": example_event,
        "client": (None, 0),
        "headers": [
            [
                b"accept",
                b"text/html,application/xhtml+xml,application/xml;"
                b"q=0.9,image/webp,image/apng,*/*;q=0.8,"
                b"application/signed-exchange;v=b3;q=0.9",
            ],
            [b"accept-encoding", b"gzip, deflate, br"],
            [b"host", b"70ixmpl4fl.execute-api.us-east-2.amazonaws.com"],
            [
                b"user-agent",
                b"Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
                b"AppleWebKit/537.36 (KHTML, like Gecko) "
                b"Chrome/80.0.3987.132 "
                b"Safari/537.36",
            ],
            [b"x-amzn-trace-id", b"Root=1-5e66d96f-7491f09xmpl79d18acf3d050"],
        ],
        "http_version": "1.1",
        "method": "GET",
        "path": "/",
        "query_string": b"foo=bar",
        "raw_path": None,
        "root_path": "",
        "scheme": "https",
        "server": ("70ixmpl4fl.execute-api.us-east-2.amazonaws.com", 80),
        "type": "http",
    }


@pytest.mark.parametrize(
    "method,path,multi_value_query_parameters,req_body,body_base64_encoded,query_string,scope_body",
    [
        ("GET", "/hello/world", None, None, False, b"", None),
        (
            "POST",
            "/",
            {"name": ["me"]},
            "field1=value1&field2=value2",
            False,
            b"name=me",
            b"field1=value1&field2=value2",
        ),
        (
            "GET",
            "/my/resource",
            {"name": ["me", "you"]},
            None,
            False,
            b"name=me&name=you",
            None,
        ),
        (
            "GET",
            "",
            {"name": ["me", "you"], "pet": ["dog"]},
            None,
            False,
            b"name=me&name=you&pet=dog",
            None,
        ),
        # A 1x1 red px gif
        (
            "POST",
            "/img",
            None,
            b"R0lGODdhAQABAIABAP8AAAAAACwAAAAAAQABAAACAkQBADs=",
            True,
            b"",
            b"GIF87a\x01\x00\x01\x00\x80\x01\x00\xff\x00\x00\x00\x00\x00,"
            b"\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02D\x01\x00;",
        ),
        ("POST", "/form-submit", None, b"say=Hi&to=Mom", False, b"", b"say=Hi&to=Mom"),
    ],
)
def test_aws_api_gateway_scope_real(
    method,
    path,
    multi_value_query_parameters,
    req_body,
    body_base64_encoded,
    query_string,
    scope_body,
):
    event = get_mock_aws_api_gateway_event(method, path, multi_value_query_parameters, req_body, body_base64_encoded)
    example_context = {}
    handler = APIGateway(event, example_context, {"api_gateway_base_path": "/"})

    scope_path = path
    if scope_path == "":
        scope_path = "/"

    assert handler.scope == {
        "asgi": {"version": "3.0", "spec_version": "2.0"},
        "aws.context": {},
        "aws.event": event,
        "client": ("192.168.100.1", 0),
        "headers": [
            [
                b"accept",
                b"text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
            ],
            [b"accept-encoding", b"gzip, deflate, lzma, sdch, br"],
            [b"accept-language", b"en-US,en;q=0.8"],
            [b"cloudfront-forwarded-proto", b"https"],
            [b"cloudfront-is-desktop-viewer", b"true"],
            [b"cloudfront-is-mobile-viewer", b"false"],
            [b"cloudfront-is-smarttv-viewer", b"false"],
            [b"cloudfront-is-tablet-viewer", b"false"],
            [b"cloudfront-viewer-country", b"US"],
            [b"cookie", b"cookie1; cookie2"],
            [b"host", b"test.execute-api.us-west-2.amazonaws.com"],
            [b"upgrade-insecure-requests", b"1"],
            [b"x-forwarded-for", b"192.168.100.1, 192.168.1.1"],
            [b"x-forwarded-port", b"443"],
            [b"x-forwarded-proto", b"https"],
        ],
        "http_version": "1.1",
        "method": method,
        "path": scope_path,
        "query_string": query_string,
        "raw_path": None,
        "root_path": "",
        "scheme": "https",
        "server": ("test.execute-api.us-west-2.amazonaws.com", 443),
        "type": "http",
    }

    if handler.body:
        assert handler.body == scope_body
    else:
        assert handler.body == b""


@pytest.mark.parametrize(
    "method,path,multi_value_query_parameters,req_body,body_base64_encoded,query_string,scope_body",
    [
        ("GET", "/test/hello", None, None, False, b"", None),
    ],
)
def test_aws_api_gateway_base_path(
    method,
    path,
    multi_value_query_parameters,
    req_body,
    body_base64_encoded,
    query_string,
    scope_body,
):
    event = get_mock_aws_api_gateway_event(method, path, multi_value_query_parameters, req_body, body_base64_encoded)

    async def app(scope, receive, send):
        assert scope["type"] == "http"
        assert scope["path"] == urllib.parse.unquote(event["path"])
        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": [[b"content-type", b"text/plain"]],
            }
        )
        await send({"type": "http.response.body", "body": b"Hello world!"})

    handler = Mangum(app, lifespan="off", api_gateway_base_path=None)
    response = handler(event, {})

    assert response == {
        "body": "Hello world!",
        "headers": {"content-type": "text/plain"},
        "multiValueHeaders": {},
        "isBase64Encoded": False,
        "statusCode": 200,
    }

    async def app(scope, receive, send):
        assert scope["type"] == "http"
        assert scope["path"] == urllib.parse.unquote(event["path"][len(f"/{api_gateway_base_path}") :])
        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": [[b"content-type", b"text/plain"]],
            }
        )
        await send({"type": "http.response.body", "body": b"Hello world!"})

    api_gateway_base_path = "test"
    handler = Mangum(app, lifespan="off", api_gateway_base_path=api_gateway_base_path)
    response = handler(event, {})
    assert response == {
        "body": "Hello world!",
        "headers": {"content-type": "text/plain"},
        "multiValueHeaders": {},
        "isBase64Encoded": False,
        "statusCode": 200,
    }


@pytest.mark.parametrize(
    "method,content_type,raw_res_body,res_body,res_base64_encoded",
    [
        ("GET", b"text/plain; charset=utf-8", b"Hello world", "Hello world", False),
        # A 1x1 red px gif
        (
            "POST",
            b"image/gif",
            b"GIF87a\x01\x00\x01\x00\x80\x01\x00\xff\x00\x00\x00\x00\x00,"
            b"\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02D\x01\x00;",
            "R0lGODdhAQABAIABAP8AAAAAACwAAAAAAQABAAACAkQBADs=",
            True,
        ),
    ],
)
def test_aws_api_gateway_response(method, content_type, raw_res_body, res_body, res_base64_encoded):
    async def app(scope, receive, send):
        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": [[b"content-type", content_type]],
            }
        )
        await send({"type": "http.response.body", "body": raw_res_body})

    event = get_mock_aws_api_gateway_event(method, "/test", {}, None, False)

    handler = Mangum(app, lifespan="off")

    response = handler(event, {})
    assert response == {
        "statusCode": 200,
        "isBase64Encoded": res_base64_encoded,
        "headers": {"content-type": content_type.decode()},
        "multiValueHeaders": {},
        "body": res_body,
    }


def test_aws_api_gateway_response_extra_mime_types():
    content_type = b"application/x-yaml"
    utf_res_body = "name: 'John Doe'"
    raw_res_body = utf_res_body.encode()
    b64_res_body = "bmFtZTogJ0pvaG4gRG9lJw=="

    async def app(scope, receive, send):
        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": [[b"content-type", content_type]],
            }
        )
        await send({"type": "http.response.body", "body": raw_res_body})

    event = get_mock_aws_api_gateway_event("POST", "/test", {}, None, False)

    # Test default behavior
    handler = Mangum(app, lifespan="off")
    response = handler(event, {})
    assert content_type.decode() not in handler.config["text_mime_types"]
    assert response == {
        "statusCode": 200,
        "isBase64Encoded": True,
        "headers": {"content-type": content_type.decode()},
        "multiValueHeaders": {},
        "body": b64_res_body,
    }

    # Test with modified text mime types
    handler = Mangum(app, lifespan="off")
    handler.config["text_mime_types"].append(content_type.decode())
    response = handler(event, {})
    assert response == {
        "statusCode": 200,
        "isBase64Encoded": False,
        "headers": {"content-type": content_type.decode()},
        "multiValueHeaders": {},
        "body": utf_res_body,
    }


def test_aws_api_gateway_exclude_headers():
    async def app(scope, receive, send):
        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": [
                    [b"content-type", b"text/plain; charset=utf-8"],
                    [b"x-custom-header", b"test"],
                ],
            }
        )
        await send({"type": "http.response.body", "body": b"Hello world"})

    event = get_mock_aws_api_gateway_event("GET", "/test", {}, None, False)

    handler = Mangum(app, lifespan="off", exclude_headers=["X-CUSTOM-HEADER"])

    response = handler(event, {})
    assert response == {
        "statusCode": 200,
        "isBase64Encoded": False,
        "headers": {"content-type": b"text/plain; charset=utf-8".decode()},
        "multiValueHeaders": {},
        "body": "Hello world",
    }



from __future__ import annotations

from typing import Any
from urllib.parse import urlencode

from mangum.handlers.utils import (
    get_server_and_port,
    handle_base64_response_body,
    handle_exclude_headers,
    handle_multi_value_headers,
    maybe_encode_body,
    strip_api_gateway_path,
)
from mangum.types import (
    Headers,
    LambdaConfig,
    LambdaContext,
    LambdaEvent,
    QueryParams,
    Response,
    Scope,
)


def _encode_query_string_for_apigw(event: LambdaEvent) -> bytes:
    params: QueryParams = event.get("multiValueQueryStringParameters", {})
    if not params:
        params = event.get("queryStringParameters", {})
    if not params:
        return b""

    return urlencode(params, doseq=True).encode()


def _handle_multi_value_headers_for_request(event: LambdaEvent) -> dict[str, str]:
    headers = event.get("headers", {}) or {}
    headers = {k.lower(): v for k, v in headers.items()}
    if event.get("multiValueHeaders"):
        headers.update(
            {
                k.lower(): ", ".join(v) if isinstance(v, list) else ""
                for k, v in event.get("multiValueHeaders", {}).items()
            }
        )

    return headers


def _combine_headers_v2(
    input_headers: Headers,
) -> tuple[dict[str, str], list[str]]:
    output_headers: dict[str, str] = {}
    cookies: list[str] = []
    for key, value in input_headers:
        normalized_key: str = key.decode().lower()
        normalized_value: str = value.decode()
        if normalized_key == "set-cookie":
            cookies.append(normalized_value)
        else:
            if normalized_key in output_headers:
                normalized_value = f"{output_headers[normalized_key]},{normalized_value}"
            output_headers[normalized_key] = normalized_value

    return output_headers, cookies


class APIGateway:
    @classmethod
    def infer(cls, event: LambdaEvent, context: LambdaContext, config: LambdaConfig) -> bool:
        return "resource" in event and "requestContext" in event

    def __init__(self, event: LambdaEvent, context: LambdaContext, config: LambdaConfig) -> None:
        self.event = event
        self.context = context
        self.config = config

    @property
    def body(self) -> bytes:
        return maybe_encode_body(
            self.event.get("body", b""),
            is_base64=self.event.get("isBase64Encoded", False),
        )

    @property
    def scope(self) -> Scope:
        headers = _handle_multi_value_headers_for_request(self.event)
        return {
            "type": "http",
            "http_version": "1.1",
            "method": self.event["httpMethod"],
            "headers": [[k.encode(), v.encode()] for k, v in headers.items()],
            "path": strip_api_gateway_path(
                self.event["path"],
                api_gateway_base_path=self.config["api_gateway_base_path"],
            ),
            "raw_path": None,
            "root_path": "",
            "scheme": headers.get("x-forwarded-proto", "https"),
            "query_string": _encode_query_string_for_apigw(self.event),
            "server": get_server_and_port(headers),
            "client": (
                self.event["requestContext"].get("identity", {}).get("sourceIp"),
                0,
            ),
            "asgi": {"version": "3.0", "spec_version": "2.0"},
            "aws.event": self.event,
            "aws.context": self.context,
        }

    def __call__(self, response: Response) -> dict[str, Any]:
        finalized_headers, multi_value_headers = handle_multi_value_headers(response["headers"])
        finalized_body, is_base64_encoded = handle_base64_response_body(
            response["body"], finalized_headers, self.config["text_mime_types"]
        )

        return {
            "statusCode": response["status"],
            "headers": handle_exclude_headers(finalized_headers, self.config),
            "multiValueHeaders": handle_exclude_headers(multi_value_headers, self.config),
            "body": finalized_body,
            "isBase64Encoded": is_base64_encoded,
        }


class HTTPGateway:
    @classmethod
    def infer(cls, event: LambdaEvent, context: LambdaContext, config: LambdaConfig) -> bool:
        return "version" in event and "requestContext" in event

    def __init__(self, event: LambdaEvent, context: LambdaContext, config: LambdaConfig) -> None:
        self.event = event
        self.context = context
        self.config = config

    @property
    def body(self) -> bytes:
        return maybe_encode_body(
            self.event.get("body", b""),
            is_base64=self.event.get("isBase64Encoded", False),
        )

    @property
    def scope(self) -> Scope:
        request_context = self.event["requestContext"]
        event_version = self.event["version"]

        # API Gateway v2
        if event_version == "2.0":
            headers = {k.lower(): v for k, v in self.event.get("headers", {}).items()}
            source_ip = request_context["http"]["sourceIp"]
            path = request_context["http"]["path"]
            http_method = request_context["http"]["method"]
            query_string = self.event.get("rawQueryString", "").encode()

            if self.event.get("cookies"):
                headers["cookie"] = "; ".join(self.event.get("cookies", []))

        # API Gateway v1
        else:
            headers = _handle_multi_value_headers_for_request(self.event)
            source_ip = request_context.get("identity", {}).get("sourceIp")
            path = self.event["path"]
            http_method = self.event["httpMethod"]
            query_string = _encode_query_string_for_apigw(self.event)

        path = strip_api_gateway_path(
            path,
            api_gateway_base_path=self.config["api_gateway_base_path"],
        )
        server = get_server_and_port(headers)
        client = (source_ip, 0)

        return {
            "type": "http",
            "method": http_method,
            "http_version": "1.1",
            "headers": [[k.encode(), v.encode()] for k, v in headers.items()],
            "path": path,
            "raw_path": None,
            "root_path": "",
            "scheme": headers.get("x-forwarded-proto", "https"),
            "query_string": query_string,
            "server": server,
            "client": client,
            "asgi": {"version": "3.0", "spec_version": "2.0"},
            "aws.event": self.event,
            "aws.context": self.context,
        }

    def __call__(self, response: Response) -> dict[str, Any]:
        if self.scope["aws.event"]["version"] == "2.0":
            finalized_headers, cookies = _combine_headers_v2(response["headers"])

            if "content-type" not in finalized_headers and response["body"] is not None:
                finalized_headers["content-type"] = "application/json"

            finalized_body, is_base64_encoded = handle_base64_response_body(
                response["body"], finalized_headers, self.config["text_mime_types"]
            )
            response_out = {
                "statusCode": response["status"],
                "body": finalized_body,
                "headers": finalized_headers or None,
                "cookies": cookies or None,
                "isBase64Encoded": is_base64_encoded,
            }
            return {key: value for key, value in response_out.items() if value is not None}

        finalized_headers, multi_value_headers = handle_multi_value_headers(response["headers"])
        finalized_body, is_base64_encoded = handle_base64_response_body(
            response["body"], finalized_headers, self.config["text_mime_types"]
        )
        return {
            "statusCode": response["status"],
            "headers": finalized_headers,
            "multiValueHeaders": multi_value_headers,
            "body": finalized_body,
            "isBase64Encoded": is_base64_encoded,
        }



import logging

import pytest
from quart import Quart
from starlette.applications import Starlette
from starlette.responses import PlainTextResponse
from typing_extensions import Literal

from mangum import Mangum
from mangum.exceptions import LifespanFailure
from mangum.types import Receive, Scope, Send


@pytest.mark.parametrize(
    "mock_aws_api_gateway_event,lifespan",
    [
        (["GET", None, None], "auto"),
        (["GET", None, None], "on"),
        (["GET", None, None], "off"),
    ],
    indirect=["mock_aws_api_gateway_event"],
)
def test_lifespan(mock_aws_api_gateway_event, lifespan) -> None:
    """
    Test each lifespan option using an application that supports lifespan messages.

    * "auto" (default):
        Application support for lifespan will be inferred.

        Any error that occurs during startup will be logged and the ASGI application
        cycle will continue unless a `lifespan.startup.failed` event is sent.

    * "on":
        Application support for lifespan is explicit.

        Any error that occurs during startup will be raised and a 500 response will
        be returned.

    * "off":
        Application support for lifespan should be ignored.

        The application will not enter the lifespan cycle context.
    """
    startup_complete = False
    shutdown_complete = False

    async def app(scope, receive, send):
        nonlocal startup_complete, shutdown_complete

        if scope["type"] == "lifespan":
            while True:
                message = await receive()
                if message["type"] == "lifespan.startup":
                    await send({"type": "lifespan.startup.complete"})
                    startup_complete = True
                elif message["type"] == "lifespan.shutdown":
                    await send({"type": "lifespan.shutdown.complete"})
                    shutdown_complete = True
                    return

        if scope["type"] == "http":
            await send(
                {
                    "type": "http.response.start",
                    "status": 200,
                    "headers": [[b"content-type", b"text/plain; charset=utf-8"]],
                }
            )
            await send({"type": "http.response.body", "body": b"Hello, world!"})

    handler = Mangum(app, lifespan=lifespan)
    response = handler(mock_aws_api_gateway_event, {})
    expected = lifespan in ("on", "auto")

    assert startup_complete == expected
    assert shutdown_complete == expected
    assert response == {
        "statusCode": 200,
        "isBase64Encoded": False,
        "headers": {"content-type": "text/plain; charset=utf-8"},
        "multiValueHeaders": {},
        "body": "Hello, world!",
    }


@pytest.mark.parametrize(
    "mock_aws_api_gateway_event,lifespan",
    [
        (["GET", None, None], "auto"),
        (["GET", None, None], "on"),
        (["GET", None, None], "off"),
    ],
    indirect=["mock_aws_api_gateway_event"],
)
def test_lifespan_unsupported(mock_aws_api_gateway_event, lifespan) -> None:
    """
    Test each lifespan option with an application that does not support lifespan events.
    """

    async def app(scope, receive, send):
        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": [[b"content-type", b"text/plain; charset=utf-8"]],
            }
        )
        await send({"type": "http.response.body", "body": b"Hello, world!"})

    handler = Mangum(app, lifespan=lifespan)
    response = handler(mock_aws_api_gateway_event, {})

    assert response == {
        "statusCode": 200,
        "isBase64Encoded": False,
        "headers": {"content-type": "text/plain; charset=utf-8"},
        "multiValueHeaders": {},
        "body": "Hello, world!",
    }


@pytest.mark.parametrize(
    "mock_aws_api_gateway_event,lifespan",
    [(["GET", None, None], "auto"), (["GET", None, None], "on")],
    indirect=["mock_aws_api_gateway_event"],
)
def test_lifespan_error(mock_aws_api_gateway_event, lifespan, caplog) -> None:
    caplog.set_level(logging.ERROR)

    async def app(scope, receive, send):
        if scope["type"] == "lifespan":
            while True:
                message = await receive()
                if message["type"] == "lifespan.startup":
                    raise Exception("error")
        else:
            await send(
                {
                    "type": "http.response.start",
                    "status": 200,
                    "headers": [[b"content-type", b"text/plain; charset=utf-8"]],
                }
            )
            await send({"type": "http.response.body", "body": b"Hello, world!"})

    handler = Mangum(app, lifespan=lifespan)
    response = handler(mock_aws_api_gateway_event, {})

    assert "Exception in 'lifespan' protocol." in caplog.text
    assert response == {
        "statusCode": 200,
        "isBase64Encoded": False,
        "headers": {"content-type": "text/plain; charset=utf-8"},
        "multiValueHeaders": {},
        "body": "Hello, world!",
    }


@pytest.mark.parametrize(
    "mock_aws_api_gateway_event,lifespan",
    [(["GET", None, None], "auto"), (["GET", None, None], "on")],
    indirect=["mock_aws_api_gateway_event"],
)
def test_lifespan_unexpected_message(mock_aws_api_gateway_event, lifespan) -> None:
    async def app(scope, receive, send):
        if scope["type"] == "lifespan":
            while True:
                message = await receive()
                if message["type"] == "lifespan.startup":
                    await send(
                        {
                            "type": "http.response.start",
                            "status": 200,
                            "headers": [[b"content-type", b"text/plain; charset=utf-8"]],
                        }
                    )

    handler = Mangum(app, lifespan=lifespan)
    with pytest.raises(LifespanFailure):
        handler(mock_aws_api_gateway_event, {})


@pytest.mark.parametrize(
    "mock_aws_api_gateway_event,lifespan,failure_type",
    [
        (["GET", None, None], "auto", "startup"),
        (["GET", None, None], "on", "startup"),
        (["GET", None, None], "auto", "shutdown"),
        (["GET", None, None], "on", "shutdown"),
    ],
    indirect=["mock_aws_api_gateway_event"],
)
def test_lifespan_failure(mock_aws_api_gateway_event, lifespan, failure_type) -> None:
    async def app(scope, receive, send):
        if scope["type"] == "lifespan":
            while True:
                message = await receive()
                if message["type"] == "lifespan.startup":
                    if failure_type == "startup":
                        await send({"type": "lifespan.startup.failed", "message": "Failed."})
                    else:
                        await send({"type": "lifespan.startup.complete"})
                if message["type"] == "lifespan.shutdown":
                    if failure_type == "shutdown":
                        await send({"type": "lifespan.shutdown.failed", "message": "Failed."})
                    await send({"type": "lifespan.shutdown.complete"})

    handler = Mangum(app, lifespan=lifespan)

    with pytest.raises(LifespanFailure):
        handler(mock_aws_api_gateway_event, {})


@pytest.mark.parametrize(
    "mock_aws_api_gateway_event,lifespan",
    [(["GET", None, None], "auto"), (["GET", None, None], "on")],
    indirect=["mock_aws_api_gateway_event"],
)
def test_lifespan_state(mock_aws_api_gateway_event, lifespan: Literal["on", "auto"]) -> None:
    startup_complete = False
    shutdown_complete = False

    async def app(scope: Scope, receive: Receive, send: Send):
        nonlocal startup_complete, shutdown_complete

        if scope["type"] == "lifespan":
            while True:
                message = await receive()
                if message["type"] == "lifespan.startup":
                    scope["state"].update({"test_key": b"Hello, world!"})
                    await send({"type": "lifespan.startup.complete"})
                    startup_complete = True
                elif message["type"] == "lifespan.shutdown":
                    await send({"type": "lifespan.shutdown.complete"})
                    shutdown_complete = True
                    return

        if scope["type"] == "http":
            await send(
                {
                    "type": "http.response.start",
                    "status": 200,
                    "headers": [[b"content-type", b"text/plain; charset=utf-8"]],
                }
            )
            await send({"type": "http.response.body", "body": scope["state"]["test_key"]})

    handler = Mangum(app, lifespan=lifespan)
    response = handler(mock_aws_api_gateway_event, {})

    assert startup_complete
    assert shutdown_complete
    assert response == {
        "statusCode": 200,
        "isBase64Encoded": False,
        "headers": {"content-type": "text/plain; charset=utf-8"},
        "multiValueHeaders": {},
        "body": "Hello, world!",
    }


@pytest.mark.parametrize("mock_aws_api_gateway_event", [["GET", None, None]], indirect=True)
def test_starlette_lifespan(mock_aws_api_gateway_event) -> None:
    startup_complete = False
    shutdown_complete = False

    path = mock_aws_api_gateway_event["path"]
    app = Starlette()

    @app.on_event("startup")
    async def on_startup():
        nonlocal startup_complete
        startup_complete = True

    @app.on_event("shutdown")
    async def on_shutdown():
        nonlocal shutdown_complete
        shutdown_complete = True

    @app.route(path)
    def homepage(request):
        return PlainTextResponse("Hello, world!")

    assert not startup_complete
    assert not shutdown_complete

    handler = Mangum(app)
    mock_aws_api_gateway_event["body"] = None

    response = handler(mock_aws_api_gateway_event, {})
    assert startup_complete
    assert shutdown_complete
    assert response == {
        "statusCode": 200,
        "isBase64Encoded": False,
        "headers": {
            "content-length": "13",
            "content-type": "text/plain; charset=utf-8",
        },
        "multiValueHeaders": {},
        "body": "Hello, world!",
    }


@pytest.mark.parametrize("mock_aws_api_gateway_event", [["GET", None, None]], indirect=True)
def test_quart_lifespan(mock_aws_api_gateway_event) -> None:
    startup_complete = False
    shutdown_complete = False
    path = mock_aws_api_gateway_event["path"]
    app = Quart(__name__)

    @app.before_serving
    async def on_startup():
        nonlocal startup_complete
        startup_complete = True

    @app.after_serving
    async def on_shutdown():
        nonlocal shutdown_complete
        shutdown_complete = True

    @app.route(path)
    async def hello():
        return "hello world!"

    assert not startup_complete
    assert not shutdown_complete

    handler = Mangum(app)
    response = handler(mock_aws_api_gateway_event, {})

    assert startup_complete
    assert shutdown_complete
    assert response == {
        "statusCode": 200,
        "isBase64Encoded": False,
        "headers": {"content-length": "12", "content-type": "text/html; charset=utf-8"},
        "multiValueHeaders": {},
        "body": "hello world!",
    }



import urllib.parse

import pytest

from mangum import Mangum
from mangum.handlers.api_gateway import HTTPGateway


def get_mock_aws_http_gateway_event_v1(method, path, query_parameters, body, body_base64_encoded):
    query_string = urllib.parse.urlencode(query_parameters if query_parameters else {})
    return {
        "version": "1.0",
        "resource": path,
        "path": path,
        "httpMethod": method,
        "headers": {
            "accept-encoding": "gzip,deflate",
            "x-forwarded-port": "443",
            "x-forwarded-proto": "https",
            "host": "test.execute-api.us-west-2.amazonaws.com",
        },
        "multiValueHeaders": {
            "accept-encoding": ["gzip", "deflate"],
            "x-forwarded-port": ["443"],
            "x-forwarded-proto": ["https"],
            "host": ["test.execute-api.us-west-2.amazonaws.com"],
        },
        "queryStringParameters": ({k: v[0] for k, v in query_parameters.items()} if query_parameters else {}),
        "multiValueQueryStringParameters": ({k: v for k, v in query_parameters.items()} if query_parameters else {}),
        "requestContext": {
            "accountId": "123456789012",
            "apiId": "id",
            "authorizer": {"claims": None, "scopes": None},
            "domainName": "id.execute-api.us-east-1.amazonaws.com",
            "domainPrefix": "id",
            "extendedRequestId": "request-id",
            "httpMethod": method,
            "identity": {
                "accessKey": None,
                "accountId": None,
                "caller": None,
                "cognitoAuthenticationProvider": None,
                "cognitoAuthenticationType": None,
                "cognitoIdentityId": None,
                "cognitoIdentityPoolId": None,
                "principalOrgId": None,
                "sourceIp": "192.168.100.1",
                "user": None,
                "userAgent": "user-agent",
                "userArn": None,
                "clientCert": {
                    "clientCertPem": "CERT_CONTENT",
                    "subjectDN": "www.example.com",
                    "issuerDN": "Example issuer",
                    "serialNumber": "a1:a1:a1:a1:a1:a1:a1:a1:a1:a1:a1:a1:a1:a1:a1:a1",
                    "validity": {
                        "notBefore": "May 28 12:30:02 2019 GMT",
                        "notAfter": "Aug  5 09:36:04 2021 GMT",
                    },
                },
            },
            "path": path,
            "protocol": "HTTP/1.1",
            "requestId": "id=",
            "requestTime": "04/Mar/2020:19:15:17 +0000",
            "requestTimeEpoch": 1583349317135,
            "resourceId": None,
            "resourcePath": path,
            "stage": "$default",
        },
        "pathParameters": query_string,
        "stageVariables": None,
        "body": body,
        "isBase64Encoded": body_base64_encoded,
    }


def get_mock_aws_http_gateway_event_v2(method, path, query_parameters, body, body_base64_encoded):
    query_string = urllib.parse.urlencode(query_parameters if query_parameters else {})
    return {
        "version": "2.0",
        "routeKey": "$default",
        "rawPath": path,
        "rawQueryString": query_string,
        "cookies": ["cookie1", "cookie2"],
        "headers": {
            "accept-encoding": "gzip,deflate",
            "x-forwarded-port": "443",
            "x-forwarded-proto": "https",
            "host": "test.execute-api.us-west-2.amazonaws.com",
        },
        "queryStringParameters": ({k: v[0] for k, v in query_parameters.items()} if query_parameters else {}),
        "requestContext": {
            "accountId": "123456789012",
            "apiId": "api-id",
            "authorizer": {
                "jwt": {
                    "claims": {"claim1": "value1", "claim2": "value2"},
                    "scopes": ["scope1", "scope2"],
                }
            },
            "domainName": "id.execute-api.us-east-1.amazonaws.com",
            "domainPrefix": "id",
            "http": {
                "method": method,
                "path": path,
                "protocol": "HTTP/1.1",
                "sourceIp": "192.168.100.1",
                "userAgent": "agent",
            },
            "requestId": "id",
            "routeKey": "$default",
            "stage": "$default",
            "time": "12/Mar/2020:19:03:58 +0000",
            "timeEpoch": 1583348638390,
        },
        "body": body,
        "pathParameters": {"parameter1": "value1"},
        "isBase64Encoded": body_base64_encoded,
        "stageVariables": {"stageVariable1": "value1", "stageVariable2": "value2"},
    }


def test_aws_http_gateway_scope_basic_v1():
    """
    Test the event from the AWS docs
    """
    example_event = {
        "version": "1.0",
        "resource": "/my/path",
        "path": "/my/path",
        "httpMethod": "GET",
        "headers": {"Header1": "value1", "Header2": "value2"},
        "multiValueHeaders": {"Header1": ["value1"], "Header2": ["value1", "value2"]},
        "queryStringParameters": {"parameter1": "value1", "parameter2": "value"},
        "multiValueQueryStringParameters": {
            "parameter1": ["value1", "value2"],
            "parameter2": ["value"],
        },
        "requestContext": {
            "accountId": "123456789012",
            "apiId": "id",
            "authorizer": {"claims": None, "scopes": None},
            "domainName": "id.execute-api.us-east-1.amazonaws.com",
            "domainPrefix": "id",
            "extendedRequestId": "request-id",
            "httpMethod": "GET",
            "identity": {
                "accessKey": None,
                "accountId": None,
                "caller": None,
                "cognitoAuthenticationProvider": None,
                "cognitoAuthenticationType": None,
                "cognitoIdentityId": None,
                "cognitoIdentityPoolId": None,
                "principalOrgId": None,
                "sourceIp": "IP",
                "user": None,
                "userAgent": "user-agent",
                "userArn": None,
                "clientCert": {
                    "clientCertPem": "CERT_CONTENT",
                    "subjectDN": "www.example.com",
                    "issuerDN": "Example issuer",
                    "serialNumber": "a1:a1:a1:a1:a1:a1:a1:a1:a1:a1:a1:a1:a1:a1:a1:a1",
                    "validity": {
                        "notBefore": "May 28 12:30:02 2019 GMT",
                        "notAfter": "Aug  5 09:36:04 2021 GMT",
                    },
                },
            },
            "path": "/my/path",
            "protocol": "HTTP/1.1",
            "requestId": "id=",
            "requestTime": "04/Mar/2020:19:15:17 +0000",
            "requestTimeEpoch": 1583349317135,
            "resourceId": None,
            "resourcePath": "/my/path",
            "stage": "$default",
        },
        "pathParameters": None,
        "stageVariables": None,
        "body": "Hello from Lambda!",
        "isBase64Encoded": False,
    }

    example_context = {}
    handler = HTTPGateway(example_event, example_context, {"api_gateway_base_path": "/"})

    assert isinstance(handler.body, bytes)
    assert handler.scope == {
        "asgi": {"version": "3.0", "spec_version": "2.0"},
        "aws.context": {},
        "aws.event": example_event,
        "client": ("IP", 0),
        "headers": [[b"header1", b"value1"], [b"header2", b"value1, value2"]],
        "http_version": "1.1",
        "method": "GET",
        "path": "/my/path",
        "query_string": b"parameter1=value1&parameter1=value2&parameter2=value",
        "raw_path": None,
        "root_path": "",
        "scheme": "https",
        "server": ("mangum", 80),
        "type": "http",
    }


def test_aws_http_gateway_scope_v1_only_non_multi_headers():
    """
    Ensure only queryStringParameters headers still works (unsure if this is possible
    from HTTP Gateway)
    """
    example_event = get_mock_aws_http_gateway_event_v1("GET", "/test", {"hello": ["world", "life"]}, None, False)
    del example_event["multiValueQueryStringParameters"]
    example_context = {}
    handler = HTTPGateway(example_event, example_context, {"api_gateway_base_path": "/"})
    assert handler.scope["query_string"] == b"hello=world"


def test_aws_http_gateway_scope_v1_no_headers():
    """
    Ensure no headers still works (unsure if this is possible from HTTP Gateway)
    """
    example_event = get_mock_aws_http_gateway_event_v1("GET", "/test", {"hello": ["world", "life"]}, None, False)
    del example_event["multiValueQueryStringParameters"]
    del example_event["queryStringParameters"]
    example_context = {}
    handler = HTTPGateway(example_event, example_context, {"api_gateway_base_path": "/"})
    assert handler.scope["query_string"] == b""


def test_aws_http_gateway_scope_basic_v2():
    """
    Test the event from the AWS docs
    """
    example_event = {
        "version": "2.0",
        "routeKey": "$default",
        "rawPath": "/my/path",
        "rawQueryString": "parameter1=value1&parameter1=value2&parameter2=value",
        "cookies": ["cookie1", "cookie2"],
        "headers": {"Header1": "value1", "Header2": "value1,value2"},
        "queryStringParameters": {"parameter1": "value1,value2", "parameter2": "value"},
        "requestContext": {
            "accountId": "123456789012",
            "apiId": "api-id",
            "authentication": {
                "clientCert": {
                    "clientCertPem": "CERT_CONTENT",
                    "subjectDN": "www.example.com",
                    "issuerDN": "Example issuer",
                    "serialNumber": "a1:a1:a1:a1:a1:a1:a1:a1:a1:a1:a1:a1:a1:a1:a1:a1",
                    "validity": {
                        "notBefore": "May 28 12:30:02 2019 GMT",
                        "notAfter": "Aug  5 09:36:04 2021 GMT",
                    },
                }
            },
            "authorizer": {
                "jwt": {
                    "claims": {"claim1": "value1", "claim2": "value2"},
                    "scopes": ["scope1", "scope2"],
                }
            },
            "domainName": "id.execute-api.us-east-1.amazonaws.com",
            "domainPrefix": "id",
            "http": {
                "method": "POST",
                "path": "/my/path",
                "protocol": "HTTP/1.1",
                "sourceIp": "IP",
                "userAgent": "agent",
            },
            "requestId": "id",
            "routeKey": "$default",
            "stage": "$default",
            "time": "12/Mar/2020:19:03:58 +0000",
            "timeEpoch": 1583348638390,
        },
        "body": "Hello from Lambda",
        "pathParameters": {"parameter1": "value1"},
        "isBase64Encoded": False,
        "stageVariables": {"stageVariable1": "value1", "stageVariable2": "value2"},
    }
    example_context = {}
    handler = HTTPGateway(example_event, example_context, {"api_gateway_base_path": "/"})

    assert isinstance(handler.body, bytes)
    assert handler.scope == {
        "asgi": {"version": "3.0", "spec_version": "2.0"},
        "aws.context": {},
        "aws.event": example_event,
        "client": ("IP", 0),
        "headers": [
            [b"header1", b"value1"],
            [b"header2", b"value1,value2"],
            [b"cookie", b"cookie1; cookie2"],
        ],
        "http_version": "1.1",
        "method": "POST",
        "path": "/my/path",
        "query_string": b"parameter1=value1&parameter1=value2&parameter2=value",
        "raw_path": None,
        "root_path": "",
        "scheme": "https",
        "server": ("mangum", 80),
        "type": "http",
    }


@pytest.mark.parametrize(
    "method,path,query_parameters,req_body,body_base64_encoded,query_string,scope_body",
    [
        ("GET", "/my/test/path", None, None, False, b"", None),
        ("GET", "", {"name": "me"}, None, False, b"name=me", None),
        # A 1x1 red px gif
        (
            "POST",
            "/img",
            None,
            b"R0lGODdhAQABAIABAP8AAAAAACwAAAAAAQABAAACAkQBADs=",
            True,
            b"",
            b"GIF87a\x01\x00\x01\x00\x80\x01\x00\xff\x00\x00\x00\x00\x00,"
            b"\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02D\x01\x00;",
        ),
        ("POST", "/form-submit", None, b"say=Hi&to=Mom", False, b"", b"say=Hi&to=Mom"),
    ],
)
def test_aws_http_gateway_scope_real_v1(
    method,
    path,
    query_parameters,
    req_body,
    body_base64_encoded,
    query_string,
    scope_body,
) -> None:
    event = get_mock_aws_http_gateway_event_v1(method, path, query_parameters, req_body, body_base64_encoded)
    example_context = {}
    handler = HTTPGateway(event, example_context, {"api_gateway_base_path": "/"})

    scope_path = path
    if scope_path == "":
        scope_path = "/"

    assert handler.scope == {
        "asgi": {"version": "3.0", "spec_version": "2.0"},
        "aws.context": {},
        "aws.event": event,
        "client": ("192.168.100.1", 0),
        "headers": [
            [b"accept-encoding", b"gzip, deflate"],
            [b"x-forwarded-port", b"443"],
            [b"x-forwarded-proto", b"https"],
            [b"host", b"test.execute-api.us-west-2.amazonaws.com"],
        ],
        "http_version": "1.1",
        "method": method,
        "path": scope_path,
        "query_string": query_string,
        "raw_path": None,
        "root_path": "",
        "scheme": "https",
        "server": ("test.execute-api.us-west-2.amazonaws.com", 443),
        "type": "http",
    }

    if handler.body:
        assert handler.body == scope_body
    else:
        assert handler.body == b""


@pytest.mark.parametrize(
    "method,path,query_parameters,req_body,body_base64_encoded,query_string,scope_body",
    [
        ("GET", "/my/test/path", None, None, False, b"", None),
        ("GET", "", {"name": "me"}, None, False, b"name=me", None),
        # A 1x1 red px gif
        (
            "POST",
            "/img",
            None,
            b"R0lGODdhAQABAIABAP8AAAAAACwAAAAAAQABAAACAkQBADs=",
            True,
            b"",
            b"GIF87a\x01\x00\x01\x00\x80\x01\x00\xff\x00\x00\x00\x00\x00,"
            b"\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02D\x01\x00;",
        ),
        ("POST", "/form-submit", None, b"say=Hi&to=Mom", False, b"", b"say=Hi&to=Mom"),
    ],
)
def test_aws_http_gateway_scope_real_v2(
    method,
    path,
    query_parameters,
    req_body,
    body_base64_encoded,
    query_string,
    scope_body,
) -> None:
    event = get_mock_aws_http_gateway_event_v2(method, path, query_parameters, req_body, body_base64_encoded)
    example_context = {}
    handler = HTTPGateway(event, example_context, {"api_gateway_base_path": "/"})

    scope_path = path
    if scope_path == "":
        scope_path = "/"

    assert handler.scope == {
        "asgi": {"version": "3.0", "spec_version": "2.0"},
        "aws.context": {},
        "aws.event": event,
        "client": ("192.168.100.1", 0),
        "headers": [
            [b"accept-encoding", b"gzip,deflate"],
            [b"x-forwarded-port", b"443"],
            [b"x-forwarded-proto", b"https"],
            [b"host", b"test.execute-api.us-west-2.amazonaws.com"],
            [b"cookie", b"cookie1; cookie2"],
        ],
        "http_version": "1.1",
        "method": method,
        "path": scope_path,
        "query_string": query_string,
        "raw_path": None,
        "root_path": "",
        "scheme": "https",
        "server": ("test.execute-api.us-west-2.amazonaws.com", 443),
        "type": "http",
    }

    if handler.body:
        assert handler.body == scope_body
    else:
        assert handler.body == b""


@pytest.mark.parametrize(
    "method,content_type,raw_res_body,res_body,res_base64_encoded",
    [
        ("GET", b"text/plain; charset=utf-8", b"Hello world", "Hello world", False),
        (
            "GET",
            b"application/json",
            b'{"hello": "world", "foo": true}',
            '{"hello": "world", "foo": true}',
            False,
        ),
        ("GET", None, b"Hello world", "SGVsbG8gd29ybGQ=", True),
        (
            "GET",
            None,
            b'{"hello": "world", "foo": true}',
            "eyJoZWxsbyI6ICJ3b3JsZCIsICJmb28iOiB0cnVlfQ==",
            True,
        ),
        # A 1x1 red px gif
        (
            "POST",
            b"image/gif",
            b"GIF87a\x01\x00\x01\x00\x80\x01\x00\xff\x00\x00\x00\x00\x00,"
            b"\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02D\x01\x00;",
            "R0lGODdhAQABAIABAP8AAAAAACwAAAAAAQABAAACAkQBADs=",
            True,
        ),
    ],
)
def test_aws_http_gateway_response_v1(method, content_type, raw_res_body, res_body, res_base64_encoded):
    """
    Test response types make sense. v1 does less magic than v2.
    """

    async def app(scope, receive, send):
        headers = []
        if content_type is not None:
            headers.append([b"content-type", content_type])

        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": headers,
            }
        )
        await send({"type": "http.response.body", "body": raw_res_body})

    event = get_mock_aws_http_gateway_event_v1(method, "/test", {}, None, False)

    handler = Mangum(app, lifespan="off")

    response = handler(event, {})

    res_headers = {}
    if content_type is not None:
        res_headers = {"content-type": content_type.decode()}

    assert response == {
        "statusCode": 200,
        "isBase64Encoded": res_base64_encoded,
        "headers": res_headers,
        "multiValueHeaders": {},
        "body": res_body,
    }


@pytest.mark.parametrize(
    "method,content_type,raw_res_body,res_body,res_base64_encoded",
    [
        ("GET", b"text/plain; charset=utf-8", b"Hello world", "Hello world", False),
        (
            "GET",
            b"application/json",
            b'{"hello": "world", "foo": true}',
            '{"hello": "world", "foo": true}',
            False,
        ),
        ("GET", None, b"Hello world", "Hello world", False),
        (
            "GET",
            None,
            b'{"hello": "world", "foo": true}',
            '{"hello": "world", "foo": true}',
            False,
        ),
        # A 1x1 red px gif
        (
            "POST",
            b"image/gif",
            b"GIF87a\x01\x00\x01\x00\x80\x01\x00\xff\x00\x00\x00\x00\x00,"
            b"\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02D\x01\x00;",
            "R0lGODdhAQABAIABAP8AAAAAACwAAAAAAQABAAACAkQBADs=",
            True,
        ),
    ],
)
def test_aws_http_gateway_response_v2(method, content_type, raw_res_body, res_body, res_base64_encoded):
    async def app(scope, receive, send):
        headers = []
        if content_type is not None:
            headers.append([b"content-type", content_type])

        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": headers,
            }
        )
        await send({"type": "http.response.body", "body": raw_res_body})

    event = get_mock_aws_http_gateway_event_v2(method, "/test", {}, None, False)

    handler = Mangum(app, lifespan="off")

    response = handler(event, {})

    if content_type is None:
        content_type = b"application/json"
    assert response == {
        "statusCode": 200,
        "isBase64Encoded": res_base64_encoded,
        "headers": {"content-type": content_type.decode()},
        "body": res_body,
    }


def test_aws_http_gateway_response_v1_extra_mime_types():
    content_type = b"application/x-yaml"
    utf_res_body = "name: 'John Doe'"
    raw_res_body = utf_res_body.encode()
    b64_res_body = "bmFtZTogJ0pvaG4gRG9lJw=="

    async def app(scope, receive, send):
        headers = []
        if content_type is not None:
            headers.append([b"content-type", content_type])

        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": headers,
            }
        )
        await send({"type": "http.response.body", "body": raw_res_body})

    event = get_mock_aws_http_gateway_event_v1("POST", "/test", {}, None, False)

    # Test default behavior
    handler = Mangum(app, lifespan="off")
    response = handler(event, {})
    assert content_type.decode() not in handler.config["text_mime_types"]
    assert response == {
        "statusCode": 200,
        "isBase64Encoded": True,
        "headers": {"content-type": content_type.decode()},
        "multiValueHeaders": {},
        "body": b64_res_body,
    }

    # Test with modified text mime types
    handler = Mangum(app, lifespan="off")
    handler.config["text_mime_types"].append(content_type.decode())
    response = handler(event, {})
    assert response == {
        "statusCode": 200,
        "isBase64Encoded": False,
        "headers": {"content-type": content_type.decode()},
        "multiValueHeaders": {},
        "body": utf_res_body,
    }


def test_aws_http_gateway_response_v2_extra_mime_types():
    content_type = b"application/x-yaml"
    utf_res_body = "name: 'John Doe'"
    raw_res_body = utf_res_body.encode()
    b64_res_body = "bmFtZTogJ0pvaG4gRG9lJw=="

    async def app(scope, receive, send):
        headers = []
        if content_type is not None:
            headers.append([b"content-type", content_type])

        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": headers,
            }
        )
        await send({"type": "http.response.body", "body": raw_res_body})

    event = get_mock_aws_http_gateway_event_v2("POST", "/test", {}, None, False)

    # Test default behavior
    handler = Mangum(app, lifespan="off")
    response = handler(event, {})
    assert content_type.decode() not in handler.config["text_mime_types"]
    assert response == {
        "statusCode": 200,
        "isBase64Encoded": True,
        "headers": {"content-type": content_type.decode()},
        "body": b64_res_body,
    }

    # Test with modified text mime types
    handler = Mangum(app, lifespan="off")
    handler.config["text_mime_types"].append(content_type.decode())
    response = handler(event, {})
    assert response == {
        "statusCode": 200,
        "isBase64Encoded": False,
        "headers": {"content-type": content_type.decode()},
        "body": utf_res_body,
    }



from __future__ import annotations

from collections.abc import Awaitable, MutableMapping, Sequence
from typing import Any, Callable, Union

from typing_extensions import Literal, Protocol, TypeAlias, TypedDict

LambdaEvent = dict[str, Any]
QueryParams: TypeAlias = MutableMapping[str, Union[str, Sequence[str]]]


class LambdaCognitoIdentity(Protocol):
    """Information about the Amazon Cognito identity that authorized the request.

    **cognito_identity_id** - The authenticated Amazon Cognito identity.
    **cognito_identity_pool_id** - The Amazon Cognito identity pool that authorized the
    invocation.
    """

    cognito_identity_id: str
    cognito_identity_pool_id: str


class LambdaMobileClient(Protocol):
    """Mobile client information for the application and the device.

    **installation_id** - A unique identifier for an installation instance of an
    application.
    **app_title** - The title of the application. For example, "My App".
    **app_version_code** - The version of the application. For example, "V2.0".
    **app_version_name** - The version code for the application. For example, 3.
    **app_package_name** - The name of the package. For example, "com.example.my_app".
    """

    installation_id: str
    app_title: str
    app_version_name: str
    app_version_code: str
    app_package_name: str


class LambdaMobileClientContext(Protocol):
    """Information about client application and device when invoked via AWS Mobile SDK.

    **client** - A dict of name-value pairs that describe the mobile client application.
    **custom** - A dict of custom values set by the mobile client application.
    **env** - A dict of environment information provided by the AWS SDK.
    """

    client: LambdaMobileClient
    custom: dict[str, Any]
    env: dict[str, Any]


class LambdaContext(Protocol):
    """The context object passed to the handler function.

    **function_name** - The name of the Lambda function.
    **function_version** - The version of the function.
    **invoked_function_arn** - The Amazon Resource Name (ARN) that's used to invoke the
    function. Indicates if the invoker specified a version number or alias.
    **memory_limit_in_mb** - The amount of memory that's allocated for the function.
    **aws_request_id** - The identifier of the invocation request.
    **log_group_name** - The log group for the function.
    **log_stream_name** - The log stream for the function instance.
    **identity** - (mobile apps) Information about the Amazon Cognito identity that
    authorized the request.
    **client_context** - (mobile apps) Client context that's provided to Lambda by the
    client application.
    """

    function_name: str
    function_version: str
    invoked_function_arn: str
    memory_limit_in_mb: int
    aws_request_id: str
    log_group_name: str
    log_stream_name: str
    identity: LambdaCognitoIdentity | None
    client_context: LambdaMobileClientContext | None

    def get_remaining_time_in_millis(self) -> int:
        """Returns the number of milliseconds left before the execution times out."""
        ...  # pragma: no cover


Headers: TypeAlias = list[list[bytes]]
Message: TypeAlias = MutableMapping[str, Any]
Scope: TypeAlias = MutableMapping[str, Any]
Receive: TypeAlias = Callable[[], Awaitable[Message]]
Send: TypeAlias = Callable[[Message], Awaitable[None]]


class ASGI(Protocol):
    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None: ...  # pragma: no cover


LifespanMode: TypeAlias = Literal["auto", "on", "off"]


class Response(TypedDict):
    status: int
    headers: Headers
    body: bytes


class LambdaConfig(TypedDict):
    api_gateway_base_path: str
    text_mime_types: list[str]
    exclude_headers: list[str]


class LambdaHandler(Protocol):
    def __init__(self, *args: Any) -> None: ...  # pragma: no cover

    @classmethod
    def infer(cls, event: LambdaEvent, context: LambdaContext, config: LambdaConfig) -> bool: ...  # pragma: no cover

    @property
    def body(self) -> bytes: ...  # pragma: no cover

    @property
    def scope(self) -> Scope: ...  # pragma: no cover

    def __call__(self, response: Response) -> dict[str, Any]: ...  # pragma: no cover



from __future__ import annotations

import asyncio
import enum
import logging
from types import TracebackType
from typing import Any

from mangum.exceptions import LifespanFailure, LifespanUnsupported, UnexpectedMessage
from mangum.types import ASGI, LifespanMode, Message


class LifespanCycleState(enum.Enum):
    """
    The state of the ASGI `lifespan` connection.

    * **CONNECTING** - Initial state. The ASGI application instance will be run with
    the connection scope containing the `lifespan` type.
    * **STARTUP** - The lifespan startup event has been pushed to the queue to be
    received by the application.
    * **SHUTDOWN** - The lifespan shutdown event has been pushed to the queue to be
    received by the application.
    * **FAILED** - A lifespan failure has been detected, and the connection will be
    closed with an error.
    * **UNSUPPORTED** - An application attempted to send a message before receiving
    the lifespan startup event. If the lifespan argument is "on", then the connection
    will be closed with an error.
    """

    CONNECTING = enum.auto()
    STARTUP = enum.auto()
    SHUTDOWN = enum.auto()
    FAILED = enum.auto()
    UNSUPPORTED = enum.auto()


class LifespanCycle:
    """
    Manages the application cycle for an ASGI `lifespan` connection.

    * **app** - An asynchronous callable that conforms to version 3.0 of the ASGI
    specification. This will usually be an ASGI framework application instance.
    * **lifespan** - A string to configure lifespan support. Choices are `auto`, `on`,
    and `off`. Default is `auto`.
    * **state** - An enumerated `LifespanCycleState` type that indicates the state of
    the ASGI connection.
    * **exception** - An exception raised while handling the ASGI event. This may or
    may not be raised depending on the state.
    * **app_queue** - An asyncio queue (FIFO) containing messages to be received by the
    application.
    * **startup_event** - An asyncio event object used to control the application
    startup flow.
    * **shutdown_event** - An asyncio event object used to control the application
    shutdown flow.
    """

    def __init__(self, app: ASGI, lifespan: LifespanMode) -> None:
        self.app = app
        self.lifespan = lifespan
        self.state: LifespanCycleState = LifespanCycleState.CONNECTING
        self.exception: BaseException | None = None
        self.loop = asyncio.get_event_loop()
        self.app_queue: asyncio.Queue[Message] = asyncio.Queue()
        self.startup_event: asyncio.Event = asyncio.Event()
        self.shutdown_event: asyncio.Event = asyncio.Event()
        self.logger = logging.getLogger("mangum.lifespan")
        self.lifespan_state: dict[str, Any] = {}

    def __enter__(self) -> None:
        """Runs the event loop for application startup."""
        self.loop.create_task(self.run())
        self.loop.run_until_complete(self.startup())

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Runs the event loop for application shutdown."""
        self.loop.run_until_complete(self.shutdown())

    async def run(self) -> None:
        """Calls the application with the `lifespan` connection scope."""
        try:
            await self.app(
                {"type": "lifespan", "asgi": {"spec_version": "2.0", "version": "3.0"}, "state": self.lifespan_state},
                self.receive,
                self.send,
            )
        except LifespanUnsupported:
            self.logger.info("ASGI 'lifespan' protocol appears unsupported.")
        except (LifespanFailure, UnexpectedMessage) as exc:
            self.exception = exc
        except BaseException as exc:
            self.logger.error("Exception in 'lifespan' protocol.", exc_info=exc)
        finally:
            self.startup_event.set()
            self.shutdown_event.set()

    async def receive(self) -> Message:
        """Awaited by the application to receive ASGI `lifespan` events."""
        if self.state is LifespanCycleState.CONNECTING:
            # Connection established. The next event returned by the queue will be
            # `lifespan.startup` to inform the application that the connection is
            # ready to receive lifespan messages.
            self.state = LifespanCycleState.STARTUP

        elif self.state is LifespanCycleState.STARTUP:
            # Connection shutting down. The next event returned by the queue will be
            # `lifespan.shutdown` to inform the application that the connection is now
            # closing so that it may perform cleanup.
            self.state = LifespanCycleState.SHUTDOWN

        return await self.app_queue.get()

    async def send(self, message: Message) -> None:
        """Awaited by the application to send ASGI `lifespan` events."""
        message_type = message["type"]
        self.logger.info("%s:  '%s' event received from application.", self.state, message_type)

        if self.state is LifespanCycleState.CONNECTING:
            if self.lifespan == "on":
                raise LifespanFailure("Lifespan connection failed during startup and lifespan is 'on'.")

            # If a message is sent before the startup event is received by the
            # application, then assume that lifespan is unsupported.
            self.state = LifespanCycleState.UNSUPPORTED
            raise LifespanUnsupported("Lifespan protocol appears unsupported.")

        if message_type not in (
            "lifespan.startup.complete",
            "lifespan.shutdown.complete",
            "lifespan.startup.failed",
            "lifespan.shutdown.failed",
        ):
            self.state = LifespanCycleState.FAILED
            raise UnexpectedMessage(f"Unexpected '{message_type}' event received.")

        if self.state is LifespanCycleState.STARTUP:
            if message_type == "lifespan.startup.complete":
                self.startup_event.set()
            elif message_type == "lifespan.startup.failed":
                self.state = LifespanCycleState.FAILED
                self.startup_event.set()
                message_value = message.get("message", "")
                raise LifespanFailure(f"Lifespan startup failure. {message_value}")

        elif self.state is LifespanCycleState.SHUTDOWN:
            if message_type == "lifespan.shutdown.complete":
                self.shutdown_event.set()
            elif message_type == "lifespan.shutdown.failed":
                self.state = LifespanCycleState.FAILED
                self.shutdown_event.set()
                message_value = message.get("message", "")
                raise LifespanFailure(f"Lifespan shutdown failure. {message_value}")

    async def startup(self) -> None:
        """Pushes the `lifespan` startup event to the queue and handles errors."""
        self.logger.info("Waiting for application startup.")
        await self.app_queue.put({"type": "lifespan.startup"})
        await self.startup_event.wait()
        if self.state is LifespanCycleState.FAILED:
            raise LifespanFailure(self.exception)

        if not self.exception:
            self.logger.info("Application startup complete.")
        else:
            self.logger.info("Application startup failed.")

    async def shutdown(self) -> None:
        """Pushes the `lifespan` shutdown event to the queue and handles errors."""
        self.logger.info("Waiting for application shutdown.")
        await self.app_queue.put({"type": "lifespan.shutdown"})
        await self.shutdown_event.wait()
        if self.state is LifespanCycleState.FAILED:
            raise LifespanFailure(self.exception)



from __future__ import annotations

from typing import Any

from mangum.types import Headers, LambdaConfig, LambdaContext, LambdaEvent, Scope


class CustomHandler:
    @classmethod
    def infer(cls, event: LambdaEvent, context: LambdaContext, config: LambdaConfig) -> bool:
        return "my-custom-key" in event

    def __init__(self, event: LambdaEvent, context: LambdaContext, config: LambdaConfig) -> None:
        self.event = event
        self.context = context
        self.config = config

    @property
    def body(self) -> bytes:
        return b"My request body"

    @property
    def scope(self) -> Scope:
        headers: dict[str, str] = {}
        return {
            "type": "http",
            "http_version": "1.1",
            "method": "GET",
            "headers": [[k.encode(), v.encode()] for k, v in headers.items()],
            "path": "/",
            "raw_path": None,
            "root_path": "",
            "scheme": "https",
            "query_string": b"",
            "server": ("mangum", 8080),
            "client": ("127.0.0.1", 0),
            "asgi": {"version": "3.0", "spec_version": "2.0"},
            "aws.event": self.event,
            "aws.context": self.context,
        }

    def __call__(self, *, status: int, headers: Headers, body: bytes) -> dict[str, Any]:
        return {"statusCode": status, "headers": {}, "body": body.decode()}


def test_custom_handler():
    event = {"my-custom-key": 1}
    handler = CustomHandler(event, {}, {"api_gateway_base_path": "/"})
    assert isinstance(handler.body, bytes)
    assert handler.scope == {
        "asgi": {"version": "3.0", "spec_version": "2.0"},
        "aws.context": {},
        "aws.event": event,
        "client": ("127.0.0.1", 0),
        "headers": [],
        "http_version": "1.1",
        "method": "GET",
        "path": "/",
        "query_string": b"",
        "raw_path": None,
        "root_path": "",
        "scheme": "https",
        "server": ("mangum", 8080),
        "type": "http",
    }



from mangum.handlers.alb import ALB
from mangum.handlers.api_gateway import APIGateway, HTTPGateway
from mangum.handlers.lambda_at_edge import LambdaAtEdge

__all__ = ["APIGateway", "HTTPGateway", "ALB", "LambdaAtEdge"]



from __future__ import annotations

import base64
from typing import Any
from urllib.parse import unquote

from mangum.types import Headers, LambdaConfig


def maybe_encode_body(body: str | bytes, *, is_base64: bool) -> bytes:
    body = body or b""
    if is_base64:
        body = base64.b64decode(body)
    elif not isinstance(body, bytes):
        body = body.encode()

    return body


def get_server_and_port(headers: dict[str, Any]) -> tuple[str, int]:
    server_name = headers.get("host", "mangum")
    if ":" not in server_name:
        server_port = headers.get("x-forwarded-port", 80)
    else:
        server_name, server_port = server_name.split(":")  # pragma: no cover
    server = (server_name, int(server_port))

    return server


def strip_api_gateway_path(path: str, *, api_gateway_base_path: str) -> str:
    if not path:
        return "/"

    if api_gateway_base_path and api_gateway_base_path != "/":
        if not api_gateway_base_path.startswith("/"):
            api_gateway_base_path = f"/{api_gateway_base_path}"
        if path.startswith(api_gateway_base_path):
            path = path[len(api_gateway_base_path) :]

    return unquote(path)


def handle_multi_value_headers(
    response_headers: Headers,
) -> tuple[dict[str, str], dict[str, list[str]]]:
    headers: dict[str, str] = {}
    multi_value_headers: dict[str, list[str]] = {}
    for key, value in response_headers:
        lower_key = key.decode().lower()
        if lower_key in multi_value_headers:
            multi_value_headers[lower_key].append(value.decode())
        elif lower_key in headers:
            # Move existing to multi_value_headers and append current
            multi_value_headers[lower_key] = [
                headers[lower_key],
                value.decode(),
            ]
            del headers[lower_key]
        else:
            headers[lower_key] = value.decode()
    return headers, multi_value_headers


def handle_base64_response_body(
    body: bytes,
    headers: dict[str, str],
    text_mime_types: list[str],
) -> tuple[str, bool]:
    is_base64_encoded = False
    output_body = ""
    if body != b"":
        for text_mime_type in text_mime_types:
            if text_mime_type in headers.get("content-type", ""):
                try:
                    output_body = body.decode()
                except UnicodeDecodeError:
                    output_body = base64.b64encode(body).decode()
                    is_base64_encoded = True
                break
        else:
            output_body = base64.b64encode(body).decode()
            is_base64_encoded = True

    return output_body, is_base64_encoded


def handle_exclude_headers(headers: dict[str, Any], config: LambdaConfig) -> dict[str, Any]:
    finalized_headers = {}
    for header_key, header_value in headers.items():
        if header_key in config["exclude_headers"]:
            continue
        finalized_headers[header_key] = header_value

    return finalized_headers



import pytest

from mangum import Mangum
from mangum.adapter import DEFAULT_TEXT_MIME_TYPES
from mangum.exceptions import ConfigurationError
from mangum.types import Receive, Scope, Send


async def app(scope: Scope, receive: Receive, send: Send): ...


def test_default_settings():
    handler = Mangum(app)
    assert handler.lifespan == "auto"
    assert handler.config["api_gateway_base_path"] == "/"
    assert sorted(handler.config["text_mime_types"]) == sorted(DEFAULT_TEXT_MIME_TYPES)
    assert handler.config["exclude_headers"] == []


@pytest.mark.parametrize(
    "arguments,message",
    [
        (
            {"lifespan": "unknown"},
            "Invalid argument supplied for `lifespan`. Choices are: auto|on|off",
        ),
    ],
)
def test_invalid_options(arguments, message):
    with pytest.raises(ConfigurationError) as exc:
        Mangum(app, **arguments)

    assert str(exc.value) == message



from __future__ import annotations

import base64
import gzip
import json
import logging
from typing import cast

import brotli
import pytest
from brotli_asgi import BrotliMiddleware
from starlette.applications import Starlette
from starlette.middleware.gzip import GZipMiddleware
from starlette.requests import Request
from starlette.responses import PlainTextResponse
from starlette.routing import Route

from mangum import Mangum


@pytest.mark.parametrize(
    "mock_aws_api_gateway_event",
    [["GET", None, {"name": ["me", "you"]}]],
    indirect=True,
)
def test_http_response(mock_aws_api_gateway_event) -> None:
    async def app(scope, receive, send):
        assert scope == {
            "asgi": {"version": "3.0", "spec_version": "2.0"},
            "aws.context": {},
            "aws.event": {
                "body": None,
                "headers": {
                    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
                    "Accept-Encoding": "gzip, deflate, lzma, sdch, br",
                    "Accept-Language": "en-US,en;q=0.8",
                    "CloudFront-Forwarded-Proto": "https",
                    "CloudFront-Is-Desktop-Viewer": "true",
                    "CloudFront-Is-Mobile-Viewer": "false",
                    "CloudFront-Is-SmartTV-Viewer": "false",
                    "CloudFront-Is-Tablet-Viewer": "false",
                    "CloudFront-Viewer-Country": "US",
                    "Cookie": "cookie1; cookie2",
                    "Host": "test.execute-api.us-west-2.amazonaws.com",
                    "Upgrade-Insecure-Requests": "1",
                    "X-Forwarded-For": "192.168.100.1, 192.168.1.1",
                    "X-Forwarded-Port": "443",
                    "X-Forwarded-Proto": "https",
                },
                "httpMethod": "GET",
                "path": "/test/hello",
                "pathParameters": {"proxy": "hello"},
                "queryStringParameters": {"name": "me"},
                "multiValueQueryStringParameters": {"name": ["me", "you"]},
                "requestContext": {
                    "accountId": "123456789012",
                    "apiId": "123",
                    "httpMethod": "GET",
                    "identity": {
                        "accountId": "",
                        "apiKey": "",
                        "caller": "",
                        "cognitoAuthenticationProvider": "",
                        "cognitoAuthenticationType": "",
                        "cognitoIdentityId": "",
                        "cognitoIdentityPoolId": "",
                        "sourceIp": "192.168.100.1",
                        "user": "",
                        "userAgent": "Mozilla/5.0 "
                        "(Macintosh; "
                        "Intel Mac OS "
                        "X 10_11_6) "
                        "AppleWebKit/537.36 "
                        "(KHTML, like "
                        "Gecko) "
                        "Chrome/52.0.2743.82 "
                        "Safari/537.36 "
                        "OPR/39.0.2256.48",
                        "userArn": "",
                    },
                    "requestId": "41b45ea3-70b5-11e6-b7bd-69b5aaebc7d9",
                    "resourceId": "us4z18",
                    "resourcePath": "/{proxy+}",
                    "stage": "Prod",
                },
                "resource": "/{proxy+}",
                "stageVariables": {"stageVarName": "stageVarValue"},
            },
            "client": ("192.168.100.1", 0),
            "headers": [
                [
                    b"accept",
                    b"text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
                ],
                [b"accept-encoding", b"gzip, deflate, lzma, sdch, br"],
                [b"accept-language", b"en-US,en;q=0.8"],
                [b"cloudfront-forwarded-proto", b"https"],
                [b"cloudfront-is-desktop-viewer", b"true"],
                [b"cloudfront-is-mobile-viewer", b"false"],
                [b"cloudfront-is-smarttv-viewer", b"false"],
                [b"cloudfront-is-tablet-viewer", b"false"],
                [b"cloudfront-viewer-country", b"US"],
                [b"cookie", b"cookie1; cookie2"],
                [b"host", b"test.execute-api.us-west-2.amazonaws.com"],
                [b"upgrade-insecure-requests", b"1"],
                [b"x-forwarded-for", b"192.168.100.1, 192.168.1.1"],
                [b"x-forwarded-port", b"443"],
                [b"x-forwarded-proto", b"https"],
            ],
            "http_version": "1.1",
            "method": "GET",
            "path": "/test/hello",
            "query_string": b"name=me&name=you",
            "raw_path": None,
            "root_path": "",
            "scheme": "https",
            "server": ("test.execute-api.us-west-2.amazonaws.com", 443),
            "type": "http",
        }
        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": [
                    [b"content-type", b"text/plain; charset=utf-8"],
                    [b"set-cookie", b"cookie1=cookie1; Secure"],
                    [b"set-cookie", b"cookie2=cookie2; Secure"],
                ],
            }
        )
        await send({"type": "http.response.body", "body": b"Hello, world!"})

    handler = Mangum(app, lifespan="off")
    response = handler(mock_aws_api_gateway_event, {})
    assert response == {
        "statusCode": 200,
        "isBase64Encoded": False,
        "headers": {"content-type": "text/plain; charset=utf-8"},
        "multiValueHeaders": {"set-cookie": ["cookie1=cookie1; Secure", "cookie2=cookie2; Secure"]},
        "body": "Hello, world!",
    }


@pytest.mark.parametrize("mock_aws_api_gateway_event", [["GET", None, None]], indirect=True)
def test_http_exception_mid_response(mock_aws_api_gateway_event) -> None:
    async def app(scope, receive, send):
        await send({"type": "http.response.start", "status": 200})
        raise Exception()

    handler = Mangum(app, lifespan="off")
    response = handler(mock_aws_api_gateway_event, {})

    assert response == {
        "body": "Internal Server Error",
        "headers": {"content-type": "text/plain; charset=utf-8"},
        "isBase64Encoded": False,
        "multiValueHeaders": {},
        "statusCode": 500,
    }


@pytest.mark.parametrize("mock_aws_api_gateway_event", [["GET", None, None]], indirect=True)
def test_http_exception_handler(mock_aws_api_gateway_event) -> None:
    path = cast(str, mock_aws_api_gateway_event["path"])

    async def all_exceptions(request: Request, exc: Exception) -> PlainTextResponse:
        return PlainTextResponse(content="Error!", status_code=500)

    def homepage(request: Request):
        raise Exception()
        return PlainTextResponse("Hello, world!")  # pragma: no cover

    app = Starlette(exception_handlers={Exception: all_exceptions}, routes=[Route(path, homepage)])

    handler = Mangum(app)
    response = handler(mock_aws_api_gateway_event, {})

    assert response == {
        "body": "Error!",
        "headers": {"content-length": "6", "content-type": "text/plain; charset=utf-8"},
        "multiValueHeaders": {},
        "isBase64Encoded": False,
        "statusCode": 500,
    }


@pytest.mark.parametrize("mock_aws_api_gateway_event", [["GET", "", None]], indirect=True)
def test_http_cycle_state(mock_aws_api_gateway_event) -> None:
    async def app(scope, receive, send):
        assert scope["type"] == "http"
        await send({"type": "http.response.body", "body": b"Hello, world!"})

    handler = Mangum(app, lifespan="off")
    response = handler(mock_aws_api_gateway_event, {})
    assert response == {
        "body": "Internal Server Error",
        "headers": {"content-type": "text/plain; charset=utf-8"},
        "multiValueHeaders": {},
        "isBase64Encoded": False,
        "statusCode": 500,
    }

    async def app(scope, receive, send):
        assert scope["type"] == "http"
        await send({"type": "http.response.start", "status": 200})
        await send({"type": "http.response.start", "status": 200})

    handler = Mangum(app, lifespan="off")

    response = handler(mock_aws_api_gateway_event, {})
    assert response == {
        "body": "Internal Server Error",
        "headers": {"content-type": "text/plain; charset=utf-8"},
        "multiValueHeaders": {},
        "isBase64Encoded": False,
        "statusCode": 500,
    }


@pytest.mark.parametrize("mock_aws_api_gateway_event", [["GET", b"", None]], indirect=True)
def test_http_binary_gzip_response(mock_aws_api_gateway_event) -> None:
    body = json.dumps({"abc": "defg"})

    async def app(scope, receive, send):
        assert scope["type"] == "http"
        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": [[b"content-type", b"application/json"]],
            }
        )

        await send({"type": "http.response.body", "body": body.encode()})

    handler = Mangum(GZipMiddleware(app, minimum_size=1), lifespan="off")
    response = handler(mock_aws_api_gateway_event, {})

    assert response["isBase64Encoded"]
    assert response["headers"] == {
        "content-encoding": "gzip",
        "content-type": "application/json",
        "content-length": "35",
        "vary": "Accept-Encoding",
    }
    # Decompress and compare content instead of comparing gzip bytes (which include timestamp)
    decompressed = gzip.decompress(base64.b64decode(response["body"]))
    assert decompressed == body.encode()


@pytest.mark.parametrize(
    "mock_http_api_event_v2",
    [
        (["GET", None, None, ""]),
        (["GET", None, {"name": ["me"]}, "name=me"]),
        (["GET", None, {"name": ["me", "you"]}, "name=me&name=you"]),
        (
            [
                "GET",
                None,
                {"name": ["me", "you"], "pet": ["dog"]},
                "name=me&name=you&pet=dog",
            ]
        ),
    ],
    indirect=["mock_http_api_event_v2"],
)
def test_set_cookies_v2(mock_http_api_event_v2) -> None:
    async def app(scope, receive, send):
        assert scope == {
            "asgi": {"version": "3.0", "spec_version": "2.0"},
            "aws.context": {},
            "aws.event": {
                "version": "2.0",
                "routeKey": "$default",
                "rawPath": "/my/path",
                "rawQueryString": mock_http_api_event_v2["rawQueryString"],
                "cookies": ["cookie1", "cookie2"],
                "headers": {
                    "accept-encoding": "gzip,deflate",
                    "x-forwarded-port": "443",
                    "x-forwarded-proto": "https",
                    "host": "test.execute-api.us-west-2.amazonaws.com",
                },
                "queryStringParameters": mock_http_api_event_v2["queryStringParameters"],
                "requestContext": {
                    "accountId": "123456789012",
                    "apiId": "api-id",
                    "authorizer": {
                        "jwt": {
                            "claims": {"claim1": "value1", "claim2": "value2"},
                            "scopes": ["scope1", "scope2"],
                        }
                    },
                    "domainName": "id.execute-api.us-east-1.amazonaws.com",
                    "domainPrefix": "id",
                    "http": {
                        "method": "GET",
                        "path": "/my/path",
                        "protocol": "HTTP/1.1",
                        "sourceIp": "192.168.100.1",
                        "userAgent": "agent",
                    },
                    "requestId": "id",
                    "routeKey": "$default",
                    "stage": "$default",
                    "time": "12/Mar/2020:19:03:58 +0000",
                    "timeEpoch": 1_583_348_638_390,
                },
                "body": None,
                "pathParameters": {"parameter1": "value1"},
                "isBase64Encoded": False,
                "stageVariables": {
                    "stageVariable1": "value1",
                    "stageVariable2": "value2",
                },
            },
            "client": ("192.168.100.1", 0),
            "headers": [
                [b"accept-encoding", b"gzip,deflate"],
                [b"x-forwarded-port", b"443"],
                [b"x-forwarded-proto", b"https"],
                [b"host", b"test.execute-api.us-west-2.amazonaws.com"],
                [b"cookie", b"cookie1; cookie2"],
            ],
            "http_version": "1.1",
            "method": "GET",
            "path": "/my/path",
            "query_string": mock_http_api_event_v2["rawQueryString"].encode(),
            "raw_path": None,
            "root_path": "",
            "scheme": "https",
            "server": ("test.execute-api.us-west-2.amazonaws.com", 443),
            "type": "http",
        }

        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": [
                    [b"content-type", b"text/plain; charset=utf-8"],
                    [b"set-cookie", b"cookie1=cookie1; Secure"],
                    [b"set-cookie", b"cookie2=cookie2; Secure"],
                    [b"multivalue", b"foo"],
                    [b"multivalue", b"bar"],
                ],
            }
        )
        await send({"type": "http.response.body", "body": b"Hello, world!"})

    handler = Mangum(app, lifespan="off")
    response = handler(mock_http_api_event_v2, {})
    assert response == {
        "statusCode": 200,
        "isBase64Encoded": False,
        "headers": {
            "content-type": "text/plain; charset=utf-8",
            "multivalue": "foo,bar",
        },
        "cookies": ["cookie1=cookie1; Secure", "cookie2=cookie2; Secure"],
        "body": "Hello, world!",
    }


@pytest.mark.parametrize(
    "mock_http_api_event_v1",
    [
        (["GET", None, None, ""]),
        (["GET", None, {"name": ["me"]}, "name=me"]),
        (["GET", None, {"name": ["me", "you"]}, "name=me&name=you"]),
        (
            [
                "GET",
                None,
                {"name": ["me", "you"], "pet": ["dog"]},
                "name=me&name=you&pet=dog",
            ]
        ),
    ],
    indirect=["mock_http_api_event_v1"],
)
def test_set_cookies_v1(mock_http_api_event_v1) -> None:
    async def app(scope, receive, send):
        assert scope == {
            "asgi": {"version": "3.0", "spec_version": "2.0"},
            "aws.context": {},
            "aws.event": {
                "version": "1.0",
                "routeKey": "$default",
                "rawPath": "/my/path",
                "path": "/my/path",
                "httpMethod": "GET",
                "rawQueryString": mock_http_api_event_v1["rawQueryString"],
                "cookies": ["cookie1", "cookie2"],
                "headers": {
                    "accept-encoding": "gzip,deflate",
                    "x-forwarded-port": "443",
                    "x-forwarded-proto": "https",
                    "host": "test.execute-api.us-west-2.amazonaws.com",
                },
                "queryStringParameters": mock_http_api_event_v1["queryStringParameters"],
                "multiValueQueryStringParameters": mock_http_api_event_v1["multiValueQueryStringParameters"],
                "requestContext": {
                    "accountId": "123456789012",
                    "apiId": "api-id",
                    "authorizer": {
                        "jwt": {
                            "claims": {"claim1": "value1", "claim2": "value2"},
                            "scopes": ["scope1", "scope2"],
                        }
                    },
                    "domainName": "id.execute-api.us-east-1.amazonaws.com",
                    "domainPrefix": "id",
                    "http": {
                        "protocol": "HTTP/1.1",
                        "sourceIp": "192.168.100.1",
                        "userAgent": "agent",
                    },
                    "requestId": "id",
                    "routeKey": "$default",
                    "stage": "$default",
                    "time": "12/Mar/2020:19:03:58 +0000",
                    "timeEpoch": 1_583_348_638_390,
                },
                "body": None,
                "pathParameters": {"parameter1": "value1"},
                "isBase64Encoded": False,
                "stageVariables": {
                    "stageVariable1": "value1",
                    "stageVariable2": "value2",
                },
            },
            "client": (None, 0),
            "headers": [
                [b"accept-encoding", b"gzip,deflate"],
                [b"x-forwarded-port", b"443"],
                [b"x-forwarded-proto", b"https"],
                [b"host", b"test.execute-api.us-west-2.amazonaws.com"],
            ],
            "http_version": "1.1",
            "method": "GET",
            "path": "/my/path",
            "query_string": mock_http_api_event_v1["rawQueryString"].encode(),
            "raw_path": None,
            "root_path": "",
            "scheme": "https",
            "server": ("test.execute-api.us-west-2.amazonaws.com", 443),
            "type": "http",
        }

        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": [
                    [b"content-type", b"text/plain; charset=utf-8"],
                    [b"set-cookie", b"cookie1=cookie1; Secure"],
                    [b"set-cookie", b"cookie2=cookie2; Secure"],
                ],
            }
        )
        await send({"type": "http.response.body", "body": b"Hello, world!"})

    handler = Mangum(app, lifespan="off")
    response = handler(mock_http_api_event_v1, {})
    assert response == {
        "statusCode": 200,
        "isBase64Encoded": False,
        "headers": {"content-type": "text/plain; charset=utf-8"},
        "multiValueHeaders": {"set-cookie": ["cookie1=cookie1; Secure", "cookie2=cookie2; Secure"]},
        "body": "Hello, world!",
    }


@pytest.mark.parametrize("mock_aws_api_gateway_event", [["GET", "", None]], indirect=True)
def test_http_empty_header(mock_aws_api_gateway_event) -> None:
    async def app(scope, receive, send):
        assert scope["type"] == "http"
        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": [[b"content-type", b"text/plain; charset=utf-8"]],
            }
        )
        await send({"type": "http.response.body", "body": b"Hello, world!"})

    handler = Mangum(app, lifespan="off")

    mock_aws_api_gateway_event["headers"] = None

    response = handler(mock_aws_api_gateway_event, {})
    assert response == {
        "statusCode": 200,
        "isBase64Encoded": False,
        "headers": {"content-type": "text/plain; charset=utf-8"},
        "multiValueHeaders": {},
        "body": "Hello, world!",
    }


@pytest.mark.parametrize(
    "mock_aws_api_gateway_event,response_headers,expected_headers,expected_multi_value_headers",
    [
        [
            ["GET", None, None],
            [[b"key1", b"value1"], [b"key2", b"value2"]],
            {"key1": "value1", "key2": "value2"},
            {},
        ],
        [
            ["GET", None, None],
            [[b"key1", b"value1"], [b"key1", b"value2"]],
            {},
            {"key1": ["value1", "value2"]},
        ],
        [
            ["GET", None, None],
            [[b"key1", b"value1"], [b"key1", b"value2"], [b"key1", b"value3"]],
            {},
            {"key1": ["value1", "value2", "value3"]},
        ],
        [["GET", None, None], [], {}, {}],
    ],
    indirect=["mock_aws_api_gateway_event"],
)
def test_http_response_headers(
    mock_aws_api_gateway_event,
    response_headers,
    expected_headers,
    expected_multi_value_headers,
):
    async def app(scope, receive, send):
        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": [[b"content-type", b"text/plain; charset=utf-8"]] + response_headers,
            }
        )
        await send({"type": "http.response.body", "body": b"Hello, world!"})

    handler = Mangum(app, lifespan="off")
    response = handler(mock_aws_api_gateway_event, {})
    expected = {
        "statusCode": 200,
        "isBase64Encoded": False,
        "headers": {"content-type": "text/plain; charset=utf-8"},
        "multiValueHeaders": {},
        "body": "Hello, world!",
    }
    if expected_headers:
        expected["headers"].update(expected_headers)
    if expected_multi_value_headers:
        expected["multiValueHeaders"] = expected_multi_value_headers
    assert response == expected


@pytest.mark.parametrize("mock_aws_api_gateway_event", [["GET", "", None]], indirect=True)
def test_http_binary_br_response(mock_aws_api_gateway_event) -> None:
    body = json.dumps({"abc": "defg"})

    async def app(scope, receive, send):
        assert scope["type"] == "http"
        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": [[b"content-type", b"application/json"]],
            }
        )

        await send({"type": "http.response.body", "body": body.encode()})

    handler = Mangum(BrotliMiddleware(app, minimum_size=1), lifespan="off")
    response = handler(mock_aws_api_gateway_event, {})

    assert response["isBase64Encoded"]
    assert response["headers"] == {
        "content-encoding": "br",
        "content-type": "application/json",
        "content-length": "19",
        "vary": "Accept-Encoding",
    }
    assert response["body"] == base64.b64encode(brotli.compress(body.encode())).decode()


@pytest.mark.parametrize("mock_aws_api_gateway_event", [["GET", b"", None]], indirect=True)
def test_http_logging(mock_aws_api_gateway_event, caplog: pytest.LogCaptureFixture) -> None:
    caplog.set_level(logging.INFO)

    async def app(scope, receive, send):
        assert scope["type"] == "http"
        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": [[b"content-type", b"text/plain; charset=utf-8"]],
            }
        )

        await send({"type": "http.response.body", "body": b"Hello, world!"})

    handler = Mangum(app, lifespan="off")
    response = handler(mock_aws_api_gateway_event, {})

    assert response == {
        "statusCode": 200,
        "isBase64Encoded": False,
        "headers": {"content-type": "text/plain; charset=utf-8"},
        "multiValueHeaders": {},
        "body": "Hello, world!",
    }

    assert "GET /test/hello 200" in caplog.text



import asyncio
import enum
import logging
from io import BytesIO

from mangum.exceptions import UnexpectedMessage
from mangum.types import ASGI, Message, Response, Scope


class HTTPCycleState(enum.Enum):
    """
    The state of the ASGI `http` connection.
    * **REQUEST** - Initial state. The ASGI application instance will be run with the
    connection scope containing the `http` type.
    * **RESPONSE** - The `http.response.start` event has been sent by the application.
    The next expected message is the `http.response.body` event, containing the body
    content. An application may pass the `more_body` argument to send content in chunks,
    however content will always be returned in a single response, never streamed.
    * **COMPLETE** - The body content from the ASGI application has been completely
    read. A disconnect event will be sent to the application, and the response will
    be returned.
    """

    REQUEST = enum.auto()
    RESPONSE = enum.auto()
    COMPLETE = enum.auto()


class HTTPCycle:
    def __init__(self, scope: Scope, body: bytes) -> None:
        self.scope = scope
        self.buffer = BytesIO()
        self.state = HTTPCycleState.REQUEST
        self.logger = logging.getLogger("mangum.http")
        self.app_queue: asyncio.Queue[Message] = asyncio.Queue()
        self.app_queue.put_nowait(
            {
                "type": "http.request",
                "body": body,
                "more_body": False,
            }
        )

    def __call__(self, app: ASGI) -> Response:
        asgi_instance = self.run(app)
        loop = asyncio.get_event_loop()
        asgi_task = loop.create_task(asgi_instance)
        loop.run_until_complete(asgi_task)

        return {
            "status": self.status,
            "headers": self.headers,
            "body": self.body,
        }

    async def run(self, app: ASGI) -> None:
        try:
            await app(self.scope, self.receive, self.send)
        except BaseException:
            self.logger.exception("An error occurred running the application.")
            if self.state is HTTPCycleState.REQUEST:
                await self.send(
                    {
                        "type": "http.response.start",
                        "status": 500,
                        "headers": [[b"content-type", b"text/plain; charset=utf-8"]],
                    }
                )
                await self.send(
                    {
                        "type": "http.response.body",
                        "body": b"Internal Server Error",
                        "more_body": False,
                    }
                )
            elif self.state is not HTTPCycleState.COMPLETE:
                self.status = 500
                self.body = b"Internal Server Error"
                self.headers = [[b"content-type", b"text/plain; charset=utf-8"]]

    async def receive(self) -> Message:
        return await self.app_queue.get()  # pragma: no cover

    async def send(self, message: Message) -> None:
        if self.state is HTTPCycleState.REQUEST and message["type"] == "http.response.start":
            self.status = message["status"]
            self.headers = message.get("headers", [])
            self.state = HTTPCycleState.RESPONSE
        elif self.state is HTTPCycleState.RESPONSE and message["type"] == "http.response.body":
            body = message.get("body", b"")
            more_body = message.get("more_body", False)
            self.buffer.write(body)
            if not more_body:
                self.body = self.buffer.getvalue()
                self.buffer.close()

                self.state = HTTPCycleState.COMPLETE
                await self.app_queue.put({"type": "http.disconnect"})

                self.logger.info(
                    "%s %s %s",
                    self.scope["method"],
                    self.scope["path"],
                    self.status,
                )
        else:
            raise UnexpectedMessage(f"Unexpected {message['type']}")
